import hashlib
import os
import sys
import tempfile
import zipfile

from humanize import naturalsize
from tqdm import tqdm
from twisted.internet import reactor
from twisted.internet.defer import inlineCallbacks
from twisted.python import log
from wormhole import __version__, create, input_with_completion

from ..errors import TransferError
from ..transit import TransitReceiver
from ..util import (bytes_to_dict, bytes_to_hexstr, dict_to_bytes,
                    estimate_free_space)
from .welcome import handle_welcome

APPID = "lothar.com/wormhole/text-or-file-xfer"

KEY_TIMER = float(os.environ.get("_MAGIC_WORMHOLE_TEST_KEY_TIMER", 1.0))
VERIFY_TIMER = float(os.environ.get("_MAGIC_WORMHOLE_TEST_VERIFY_TIMER", 1.0))


class RespondError(Exception):
    def __init__(self, response):
        self.response = response


class TransferRejectedError(RespondError):
    def __init__(self):
        RespondError.__init__(self, "transfer rejected")


def receive(args, reactor=reactor, _debug_stash_wormhole=None):
    """I implement 'wormhole receive'. I return a Deferred that fires with
    None (for success), or signals one of the following errors:
    * WrongPasswordError: the two sides didn't use matching passwords
    * Timeout: something didn't happen fast enough for our tastes
    * TransferError: the sender rejected the transfer: verifier mismatch
    * any other error: something unexpected happened
    """
    r = Receiver(args, reactor)
    d = r.go()
    if _debug_stash_wormhole is not None:
        _debug_stash_wormhole.append(r._w)
    return d


class Receiver:
    def __init__(self, args, reactor=reactor):
        assert isinstance(args.relay_url, str)
        self.args = args
        self._reactor = reactor
        self._tor = None
        self._transit_receiver = None

    def _msg(self, *args, **kwargs):
        print(*args, file=self.args.stderr, **kwargs)

    @inlineCallbacks
    def go(self):
        if self.args.tor:
            with self.args.timing.add("import", which="tor_manager"):
                from ..tor_manager import get_tor
            # For now, block everything until Tor has started. Soon: launch
            # tor in parallel with everything else, make sure the Tor object
            # can lazy-provide an endpoint, and overlap the startup process
            # with the user handing off the wormhole code
            self._tor = yield get_tor(
                self._reactor,
                self.args.launch_tor,
                self.args.tor_control_port,
                timing=self.args.timing)

        w = create(
            self.args.appid or APPID,
            self.args.relay_url,
            self._reactor,
            tor=self._tor,
            timing=self.args.timing,
        )
        if self.args.debug_state:
            w.debug_set_trace("recv", which=" ".join(self.args.debug_state), file=self.args.stdout)
        self._w = w  # so tests can wait on events too

        # I wanted to do this instead:
        #
        #    try:
        #        yield self._go(w, tor)
        #    finally:
        #        yield w.close()
        #
        # but when _go had a UsageError, the stacktrace was always displayed
        # as coming from the "yield self._go" line, which wasn't very useful
        # for tracking it down.
        d = self._go(w)

        # if we succeed, we should close and return the w.close results
        # (which might be an error)
        @inlineCallbacks
        def _good(res):
            yield w.close()  # wait for ack
            return res

        # if we raise an error, we should close and then return the original
        # error (the close might give us an error, but it isn't as important
        # as the original one)
        @inlineCallbacks
        def _bad(f):
            try:
                yield w.close()  # might be an error too
            except Exception:
                pass
            return f

        d.addCallbacks(_good, _bad)
        yield d

    @inlineCallbacks
    def _go(self, w):
        welcome = yield w.get_welcome()
        handle_welcome(welcome, self.args.relay_url, __version__,
                       self.args.stderr)

        yield self._handle_code(w)

        def on_slow_key():
            print("Waiting for sender...", file=self.args.stderr)

        notify = self._reactor.callLater(KEY_TIMER, on_slow_key)
        try:
            # We wait here until we connect to the server and see the senders
            # PAKE message. If we used set_code() in the "human-selected
            # offline codes" mode, then the sender might not have even
            # started yet, so we might be sitting here for a while. Because
            # of that possibility, it's probably not appropriate to give up
            # automatically after some timeout. The user can express their
            # impatience by quitting the program with control-C.
            yield w.get_unverified_key()
        finally:
            if not notify.called:
                notify.cancel()

        def on_slow_verification():
            print(
                "Key established, waiting for confirmation...",
                file=self.args.stderr)

        notify = self._reactor.callLater(VERIFY_TIMER, on_slow_verification)
        try:
            # We wait here until we've seen their VERSION message (which they
            # send after seeing our PAKE message, and has the side-effect of
            # verifying that we both share the same key). There is a
            # round-trip between these two events, and we could experience a
            # significant delay here if:
            # * the relay server is being restarted
            # * the network is very slow
            # * the sender is very slow
            # * the sender has quit (in which case we may wait forever)

            # It would be reasonable to give up after waiting here for too
            # long.
            verifier_bytes = yield w.get_verifier()
        finally:
            if not notify.called:
                notify.cancel()
        self._show_verifier(verifier_bytes)

        want_offer = True

        while True:
            them_d = yield self._get_data(w)
            recognized = False
            if "transit" in them_d:
                recognized = True
                yield self._parse_transit(them_d["transit"], w)
            if "offer" in them_d:
                recognized = True
                if not want_offer:
                    raise TransferError("duplicate offer")
                want_offer = False
                try:
                    yield self._parse_offer(them_d["offer"], w)
                except RespondError as r:
                    self._send_data({"error": r.response}, w)
                    raise TransferError(r.response)
                return None
            if not recognized:
                log.msg(f"unrecognized message {them_d!r}")

    def _send_data(self, data, w):
        data_bytes = dict_to_bytes(data)
        w.send_message(data_bytes)

    @inlineCallbacks
    def _get_data(self, w):
        # this may raise WrongPasswordError
        them_bytes = yield w.get_message()
        them_d = bytes_to_dict(them_bytes)
        if "error" in them_d:
            raise TransferError(them_d["error"])
        return them_d

    @inlineCallbacks
    def _handle_code(self, w):
        code = self.args.code
        if self.args.zeromode:
            assert not code
            code = "0-"
        if code:
            w.set_code(code)
        else:
            if self.args.allocate:
                w.allocate_code(self.args.code_length)
                code = yield w.get_code()
                print(f"Allocated code: {code}", file=self.args.stderr)
                print("On the other computer, please run:", file=self.args.stderr)
                print(f"   wormhole send --code {code} <filename>", file=self.args.stderr)

            else:
                prompt = "Enter receive wormhole code: "
                used_completion = yield input_with_completion(
                    prompt, w.input_code(), self._reactor)
                if not used_completion:
                    print(
                        " (note: you can use <Tab> to complete words)",
                        file=self.args.stderr)
        yield w.get_code()

    def _show_verifier(self, verifier_bytes):
        verifier_hex = bytes_to_hexstr(verifier_bytes)
        if self.args.verify:
            self._msg(f"Verifier {verifier_hex}.")

    @inlineCallbacks
    def _parse_transit(self, sender_transit, w):
        if self._transit_receiver:
            # TODO: accept multiple messages, add the additional hints to the
            # existing TransitReceiver
            return
        yield self._build_transit(w, sender_transit)

    @inlineCallbacks
    def _build_transit(self, w, sender_transit):
        tr = TransitReceiver(
            self.args.transit_helper,
            no_listen=(not self.args.listen),
            tor=self._tor,
            reactor=self._reactor,
            timing=self.args.timing)
        self._transit_receiver = tr
        # When I made it possible to override APPID with a CLI argument
        # (issue #113), I forgot to also change this w.derive_key() (issue
        # #339). We're stuck with it now. Use a local constant to make this
        # clear.
        BUG339_APPID = "lothar.com/wormhole/text-or-file-xfer"
        transit_key = w.derive_key(BUG339_APPID + "/transit-key",
                                   tr.TRANSIT_KEY_LENGTH)
        tr.set_transit_key(transit_key)

        tr.add_connection_hints(sender_transit.get("hints-v1", []))
        receiver_abilities = tr.get_connection_abilities()
        receiver_hints = yield tr.get_connection_hints()
        receiver_transit = {
            "abilities-v1": receiver_abilities,
            "hints-v1": receiver_hints,
        }
        self._send_data({"transit": receiver_transit}, w)
        # TODO: send more hints as the TransitReceiver produces them

    @inlineCallbacks
    def _parse_offer(self, them_d, w):
        if "message" in them_d:
            self._handle_text(them_d, w)
            return None
        # transit will be created by this point, but not connected
        if "file" in them_d:
            f = self._handle_file(them_d)
            self._send_permission(w)
            rp = yield self._establish_transit()
            datahash = yield self._transfer_data(rp, f)
            self._write_file(f)
            yield self._close_transit(rp, datahash)
        elif "directory" in them_d:
            f = self._handle_directory(them_d)
            self._send_permission(w)
            rp = yield self._establish_transit()
            datahash = yield self._transfer_data(rp, f)
            self._write_directory(f)
            yield self._close_transit(rp, datahash)
        else:
            self._msg("I don't know what they're offering\n")
            self._msg(f"Offer details: {them_d!r}")
            raise RespondError("unknown offer type")

    def _handle_text(self, them_d, w):
        # we're receiving a text message
        self.args.timing.add("print")
        print(repr(them_d["message"])[1:-1], file=self.args.stdout)
        self._send_data({"answer": {"message_ack": "ok"}}, w)

    def _handle_file(self, them_d):
        file_data = them_d["file"]
        self.abs_destname = self._decide_destname("file",
                                                  file_data["filename"])
        self.xfersize = file_data["filesize"]
        free = estimate_free_space(self.abs_destname)
        if free is not None and free < self.xfersize:
            self._msg("Error: insufficient free space (%sB) for file (%sB)" %
                      (free, self.xfersize))
            raise TransferRejectedError()

        # note the repr() here is (at least partially) to guard
        # against malicious filenames that might play with how
        # terminals display things. see also
        # e.g. https://github.com/magic-wormhole/magic-wormhole/issues/476
        self._msg("Receiving file (%s) into: %s" %
                  (naturalsize(self.xfersize),
                   repr(os.path.basename(self.abs_destname))))
        self._ask_permission()
        tmp_destname = self.abs_destname + ".tmp"
        return open(tmp_destname, "wb")

    def _handle_directory(self, them_d):
        file_data = them_d["directory"]
        zipmode = file_data["mode"]
        if not zipmode.startswith("zipfile"):
            self._msg(f"Error: unknown directory-transfer mode '{zipmode}'")
            raise RespondError("unknown mode")
        self.abs_destname = self._decide_destname("directory",
                                                  file_data["dirname"])
        self.xfersize = file_data["zipsize"]
        free = estimate_free_space(self.abs_destname)
        if free is not None and free < file_data["numbytes"]:
            self._msg(
                "Error: insufficient free space (%sB) for directory (%sB)" %
                (free, file_data["numbytes"]))
            raise TransferRejectedError()

        self._msg("Receiving directory (%s) into: %s/" %
                  (naturalsize(self.xfersize),
                   repr(os.path.basename(self.abs_destname))))
        self._msg("%d files, %s (uncompressed)" %
                  (file_data["numfiles"], naturalsize(file_data["numbytes"])))
        self._ask_permission()
        # max_size here matches the magic-number in cmd_send and will
        # use up to 10MB of memory before putting the file on disk
        # instead.
        f = tempfile.SpooledTemporaryFile(max_size=10*1000*1000)
        # workaround for https://bugs.python.org/issue26175 (STF doesn't
        # fully implement IOBase abstract class), which breaks the new
        # zipfile in py3.7.0 that expects .seekable
        if not hasattr(f, "seekable"):
            # AFAICT all the filetypes that STF wraps can seek
            f.seekable = lambda: True
        return f

    def _decide_destname(self, mode, destname):
        """
        Resolve any user options (--output-file) and convert to an
        absolute destination path.
        """
        # the basename() is intended to protect us against
        # "~/.ssh/authorized_keys" and other attacks
        if self.args.output_file:
            abs_destname = os.path.abspath(os.path.join(self.args.cwd, self.args.output_file))
        else:
            abs_destname = os.path.abspath(os.path.join(self.args.cwd, os.path.basename(destname)))
        overwrite_allowed = False
        if self.args.output_file:
             if os.path.exists(abs_destname):
                if os.path.isdir(abs_destname):
                    # '--output-file' is an existing directory so put
                    # the incoming file _inside_ of it (i.e.. don't
                    # delete + overwrite whole tree)
                    abs_destname = os.path.abspath(
                        os.path.join(self.args.cwd, self.args.output_file, os.path.basename(destname))
                    )
                    overwrite_allowed = True
                else:
                    # the user specified an existing _file_, so we can
                    # overwrite it
                    overwrite_allowed = True

        # get confirmation from the user before writing to the local directory
        if os.path.exists(abs_destname):
            if overwrite_allowed:  # overwrite is intentional
                self._msg(f"Overwriting {repr(self.args.output_file)}")
                if self.args.accept_file:
                    self._remove_existing(abs_destname)
            else:
                self._msg(
                    f"Error: refusing to overwrite existing {repr(destname)}")
                raise TransferRejectedError()
        return abs_destname

    def _remove_existing(self, path):
        if os.path.isfile(path):
            os.remove(path)
        if os.path.isdir(path):
            self._msg(f"Not deleting existing directory: {path}")
            raise TransferRejectedError()

    def _ask_permission(self):
        with self.args.timing.add("permission", waiting="user") as t:
            while True and not self.args.accept_file:
                ok = input("ok? (Y/n): ")
                if ok.lower().startswith("y") or len(ok) == 0:
                    if os.path.exists(self.abs_destname):
                        self._remove_existing(self.abs_destname)
                    break
                print("transfer rejected", file=sys.stderr)
                t.detail(answer="no")
                raise TransferRejectedError()
            t.detail(answer="yes")

    def _send_permission(self, w):
        self._send_data({"answer": {"file_ack": "ok"}}, w)

    @inlineCallbacks
    def _establish_transit(self):
        record_pipe = yield self._transit_receiver.connect()
        self.args.timing.add("transit connected")
        return record_pipe

    @inlineCallbacks
    def _transfer_data(self, record_pipe, f):
        # now receive the rest of the owl
        self._msg(f"Receiving ({record_pipe.describe()})..")

        with self.args.timing.add("rx file"):
            progress = tqdm(
                file=self.args.stderr,
                disable=self.args.hide_progress,
                unit="B",
                unit_scale=True,
                dynamic_ncols=True,
                total=self.xfersize)
            hasher = hashlib.sha256()
            with progress:
                received = yield record_pipe.writeToFile(
                    f, self.xfersize, progress.update, hasher.update)
            datahash = hasher.digest()

        # except TransitError
        if received < self.xfersize:
            self._msg()
            self._msg("Connection dropped before full file received")
            self._msg("got %d bytes, wanted %d" % (received, self.xfersize))
            raise TransferError("Connection dropped before full file received")
        assert received == self.xfersize
        return datahash

    def _write_file(self, f):
        tmp_name = f.name
        f.close()
        os.rename(tmp_name, self.abs_destname)
        self._msg(f"Received file written to: {self.abs_destname}")

    def _extract_file(self, zf, info, extract_dir):
        """
        the zipfile module does not restore file permissions
        so we'll do it manually
        """
        out_path = os.path.join(extract_dir, info.filename)
        out_path = os.path.abspath(out_path)
        if not out_path.startswith(extract_dir + os.sep):
            raise ValueError(
                "malicious zipfile, %s outside of extract_dir %s" %
                (info.filename, extract_dir))

        zf.extract(info.filename, path=extract_dir)

        # not sure why zipfiles store the perms 16 bits away but they do
        perm = info.external_attr >> 16
        os.chmod(out_path, perm)

    def _write_directory(self, f):
        self._msg("Unpacking zipfile..")
        with self.args.timing.add("unpack zip"):
            with zipfile.ZipFile(f, "r") as zf:
                for info in zf.infolist():
                    self._extract_file(zf, info, self.abs_destname)

            self._msg(f"Received files written to: {self.abs_destname}")
            f.close()

    @inlineCallbacks
    def _close_transit(self, record_pipe, datahash):
        datahash_hex = bytes_to_hexstr(datahash)
        ack = {"ack": "ok", "sha256": datahash_hex}
        ack_bytes = dict_to_bytes(ack)
        with self.args.timing.add("send ack"):
            yield record_pipe.send_record(ack_bytes)
            yield record_pipe.close()



# No unicode_literals
import json
import os
import unicodedata
from binascii import hexlify, unhexlify
from cryptography.hazmat.primitives.kdf import hkdf
from cryptography.hazmat.primitives import hashes
from attr import attrs, attrib


def HKDF(skm, outlen, salt=None, CTXinfo=b""):
    """
    Return the RFC5869 'HMAC-based Key Derivation Function' result of
    using the given `salt`, tag from `CTXinfo` and secret from `skm`
    with a SHA256 hash.

    :param bytes skm: the input key material
    :param int outlen: output length, in bytes
    :param bytes salt: optional salt value (None for no salt)
    :param bytes CTXinfo: context / application-specific string (default is empty)

    :return bytes: the derived key material
    """
    return hkdf.HKDF(
        hashes.SHA256(),
        outlen,
        salt,
        CTXinfo,
    ).derive(skm)


def to_bytes(u):
    return unicodedata.normalize("NFC", u).encode("utf-8")


def to_unicode(any):
    if isinstance(any, str):
        return any
    return any.decode("ascii")


def bytes_to_hexstr(b):
    assert isinstance(b, bytes)
    hexstr = hexlify(b).decode("ascii")
    assert isinstance(hexstr, str)
    return hexstr


def hexstr_to_bytes(hexstr):
    assert isinstance(hexstr, str)
    b = unhexlify(hexstr.encode("ascii"))
    assert isinstance(b, bytes)
    return b


def dict_to_bytes(d):
    assert isinstance(d, dict)
    b = json.dumps(d).encode("utf-8")
    assert isinstance(b, bytes)
    return b


def bytes_to_dict(b):
    assert isinstance(b, bytes)
    d = json.loads(b.decode("utf-8"))
    assert isinstance(d, dict)
    return d


def estimate_free_space(target):
    # f_bfree is the blocks available to a root user. It might be more
    # accurate to use f_bavail (blocks available to non-root user), but we
    # don't know which user is running us, and a lot of installations don't
    # bother with reserving extra space for root, so let's just stick to the
    # basic (larger) estimate.
    try:
        s = os.statvfs(os.path.dirname(os.path.abspath(target)))
        return s.f_frsize * s.f_bfree
    except AttributeError:
        return None


@attrs(repr=False, slots=True, hash=True)
class _ProvidesValidator:
    interface = attrib()

    def __call__(self, inst, attr, value):
        """
        We use a callable class to be able to change the ``__repr__``.
        """
        if not self.interface.providedBy(value):
            msg = "'{name}' must provide {interface!r} which {value!r} doesn't.".format(
                name=attr.name, interface=self.interface, value=value
            )
            raise TypeError(
                msg,
                attr,
                self.interface,
                value,
            )

    def __repr__(self):
        return f"<provides validator for interface {self.interface!r}>"


def provides(interface):
    """
    A validator that raises a `TypeError` if the initializer is called
    with an object that does not provide the requested *interface* (checks are
    performed using ``interface.providedBy(value)``.

    :param interface: The interface to check for.
    :type interface: ``zope.interface.Interface``

    :raises TypeError: With a human readable error message, the attribute
        (of type `attrs.Attribute`), the expected interface, and the
        value it got.
    """
    return _ProvidesValidator(interface)



import re

from automat import MethodicalMachine
from zope.interface import implementer

from . import _interfaces
from ._wordlist import PGPWordList
from .errors import KeyFormatError
from ._status import AllocatedCode, ConsumedCode


def validate_nameplate(nameplate):
    if not re.search(r'^\d+$', nameplate):
        raise KeyFormatError(
            f"Nameplate '{nameplate}' must be numeric, with no spaces.")


@implementer(_interfaces.INameplate)
class Nameplate:
    m = MethodicalMachine()
    set_trace = getattr(m, "_setTrace",
                        lambda self, f: None)  # pragma: no cover

    def __init__(self, evolve_wormhole_status):
        self._nameplate = None
        self._evolve_wormhole_status = evolve_wormhole_status

    def wire(self, mailbox, input, rendezvous_connector, terminator):
        self._M = _interfaces.IMailbox(mailbox)
        self._I = _interfaces.IInput(input)
        self._RC = _interfaces.IRendezvousConnector(rendezvous_connector)
        self._T = _interfaces.ITerminator(terminator)

    # all -A states: not connected
    # all -B states: yes connected
    # B states serialize as A, so they deserialize as unconnected

    # S0: know nothing
    @m.state(initial=True)
    def S0A(self):
        pass  # pragma: no cover

    @m.state()
    def S0B(self):
        pass  # pragma: no cover

    # S1: nameplate known, never claimed
    @m.state()
    def S1A(self):
        pass  # pragma: no cover

    # S2: nameplate known, maybe claimed
    @m.state()
    def S2A(self):
        pass  # pragma: no cover

    @m.state()
    def S2B(self):
        pass  # pragma: no cover

    # S3: nameplate claimed
    @m.state()
    def S3A(self):
        pass  # pragma: no cover

    @m.state()
    def S3B(self):
        pass  # pragma: no cover

    # S4: maybe released
    @m.state()
    def S4A(self):
        pass  # pragma: no cover

    @m.state()
    def S4B(self):
        pass  # pragma: no cover

    # S5: released
    # we no longer care whether we're connected or not
    # @m.state()
    # def S5A(self): pass
    # @m.state()
    # def S5B(self): pass
    @m.state()
    def S5(self):
        pass  # pragma: no cover

    S5A = S5
    S5B = S5

    # from Boss
    def set_nameplate(self, nameplate):
        validate_nameplate(nameplate)  # can raise KeyFormatError
        self._set_nameplate(nameplate)

    @m.input()
    def _set_nameplate(self, nameplate):
        pass

    # from Mailbox
    @m.input()
    def release(self):
        pass

    # from Terminator
    @m.input()
    def close(self):
        pass

    # from RendezvousConnector
    @m.input()
    def connected(self):
        pass

    @m.input()
    def lost(self):
        pass

    @m.input()
    def rx_claimed(self, mailbox):
        pass

    @m.input()
    def rx_released(self):
        pass

    @m.output()
    def record_nameplate(self, nameplate):
        validate_nameplate(nameplate)
        self._nameplate = nameplate

    @m.output()
    def record_nameplate_and_RC_tx_claim(self, nameplate):
        validate_nameplate(nameplate)
        self._nameplate = nameplate
        self._RC.tx_claim(self._nameplate)

    @m.output()
    def RC_tx_claim(self):
        # when invoked via M.connected(), we must use the stored nameplate
        self._RC.tx_claim(self._nameplate)

    @m.output()
    def I_got_wordlist(self, mailbox):
        # TODO select wordlist based on nameplate properties, in rx_claimed
        wordlist = PGPWordList()
        self._I.got_wordlist(wordlist)

    @m.output()
    def M_got_mailbox(self, mailbox):
        self._M.got_mailbox(mailbox)

    @m.output()
    def RC_tx_release(self):
        assert self._nameplate
        self._RC.tx_release(self._nameplate)

    @m.output()
    def T_nameplate_done(self):
        self._T.nameplate_done()

    @m.output()
    def send_status_code_allocated(self):
        self._evolve_wormhole_status(
            code=AllocatedCode(),
        )

    @m.output()
    def send_status_code_consumed(self):
        self._evolve_wormhole_status(
            code=ConsumedCode(),
        )

    S0A.upon(_set_nameplate, enter=S1A, outputs=[record_nameplate, send_status_code_allocated])
    S0A.upon(connected, enter=S0B, outputs=[])
    S0A.upon(close, enter=S5A, outputs=[T_nameplate_done, send_status_code_consumed])
    S0B.upon(
        _set_nameplate, enter=S2B, outputs=[record_nameplate_and_RC_tx_claim, send_status_code_allocated])
    S0B.upon(lost, enter=S0A, outputs=[])
    S0B.upon(close, enter=S5A, outputs=[T_nameplate_done, send_status_code_consumed])

    S1A.upon(connected, enter=S2B, outputs=[RC_tx_claim])
    S1A.upon(close, enter=S5A, outputs=[T_nameplate_done, send_status_code_consumed])

    S2A.upon(connected, enter=S2B, outputs=[RC_tx_claim])
    S2A.upon(close, enter=S4A, outputs=[])
    S2B.upon(lost, enter=S2A, outputs=[])
    S2B.upon(rx_claimed, enter=S3B, outputs=[I_got_wordlist, M_got_mailbox])
    S2B.upon(close, enter=S4B, outputs=[RC_tx_release])

    S3A.upon(connected, enter=S3B, outputs=[])
    S3A.upon(close, enter=S4A, outputs=[])
    S3B.upon(lost, enter=S3A, outputs=[])
    # S3B.upon(rx_claimed, enter=S3B, outputs=[]) # shouldn't happen
    S3B.upon(release, enter=S4B, outputs=[RC_tx_release])
    S3B.upon(close, enter=S4B, outputs=[RC_tx_release])

    S4A.upon(connected, enter=S4B, outputs=[RC_tx_release])
    S4A.upon(close, enter=S4A, outputs=[])
    S4B.upon(lost, enter=S4A, outputs=[])
    S4B.upon(rx_claimed, enter=S4B, outputs=[])
    S4B.upon(rx_released, enter=S5B, outputs=[T_nameplate_done, send_status_code_consumed])
    S4B.upon(release, enter=S4B, outputs=[])  # mailbox is lazy
    # Mailbox doesn't remember how many times it's sent a release, and will
    # re-send a new one for each peer message it receives. Ignoring it here
    # is easier than adding a new pair of states to Mailbox.
    S4B.upon(close, enter=S4B, outputs=[])

    S5A.upon(connected, enter=S5B, outputs=[])
    S5B.upon(lost, enter=S5A, outputs=[])
    S5.upon(release, enter=S5, outputs=[])  # mailbox is lazy
    S5.upon(close, enter=S5, outputs=[])



import builtins
import io
import os
import re
import stat
import sys
import tempfile
import zipfile
from functools import partial
from textwrap import dedent, fill
from pathlib import Path

from click import UsageError
from click.testing import CliRunner
from humanize import naturalsize
from twisted.internet import endpoints, reactor
from twisted.internet.defer import gatherResults, CancelledError, ensureDeferred
from twisted.internet.error import ConnectionRefusedError
from twisted.internet.utils import getProcessOutputAndValue
from twisted.python import log, procutils
from zope.interface import implementer

from unittest import mock

import pytest
import pytest_twisted

from hypothesis import given
from hypothesis import strategies as st

from .. import __version__
from .._interfaces import ITorManager
from ..cli import cli, cmd_receive, cmd_send, welcome
from ..errors import (ServerConnectionError, TransferError,
                      UnsendableFileError, WelcomeError, WrongPasswordError)
from .common import config, setup_mailbox


def build_offer(args):
    s = cmd_send.Sender(args, None)
    return s._build_offer()


def maybe_delete(fn):
    if os.path.exists(fn):
        os.unlink(fn)


def create_config():
    cfg = config("send")
    cfg.stdout = io.StringIO()
    cfg.stderr = io.StringIO()
    return cfg


def test_text_offer():
    cfg = create_config()
    cfg.text = message = "blah blah blah ponies"
    d, fd_to_send = build_offer(cfg)

    assert "message" in d
    assert "file" not in d
    assert "directory" not in d
    assert d["message"] == message
    assert fd_to_send is None


def test_file_offer(tmpdir_factory):
    cfg = create_config()
    cfg.what = filename = "my file"
    message = b"yay ponies\n"
    send_dir = tmpdir_factory.mktemp("sendfile")
    abs_filename = os.path.join(send_dir, filename)
    with open(abs_filename, "wb") as f:
        f.write(message)

    cfg.cwd = send_dir
    d, fd_to_send = build_offer(cfg)

    assert "message" not in d
    assert "file" in d
    assert "directory" not in d
    assert d["file"]["filesize"] == len(message)
    assert d["file"]["filename"] == filename
    assert fd_to_send.tell() == 0
    assert fd_to_send.read() == message


@pytest.mark.skipif(not hasattr(os, 'symlink'), reason="host OS does not support symlinks")
def _create_broken_symlink(cfg, parent_dir):
    send_dir = "dirname"
    os.mkdir(os.path.join(parent_dir, send_dir))
    os.symlink('/non/existent/file',
               os.path.join(parent_dir, send_dir, 'linky'))

    send_dir_arg = send_dir
    cfg.what = send_dir_arg
    cfg.cwd = parent_dir


def test_broken_symlink_raises_err(tmpdir_factory):
    cfg = create_config()
    _create_broken_symlink(cfg, tmpdir_factory.mktemp("broken_sym"))
    cfg.ignore_unsendable_files = False
    with pytest.raises(UnsendableFileError) as e:
        build_offer(cfg)

    # On english distributions of Linux, this will be
    # "linky: No such file or directory", but the error may be
    # different on Windows and other locales and/or Unix variants, so
    # we'll just assert the part we know about.
    assert "linky: " in str(e)


def test_broken_symlink_is_ignored(tmpdir_factory):
    cfg = create_config()
    _create_broken_symlink(cfg, tmpdir_factory.mktemp("broken_sym_ign"))
    cfg.ignore_unsendable_files = True
    d, fd_to_send = build_offer(cfg)
    assert '(ignoring error)' in cfg.stderr.getvalue()
    assert d['directory']['numfiles'] == 0
    assert d['directory']['numbytes'] == 0


def test_missing_file(tmpdir_factory):
    cfg = create_config()
    cfg.what = filename = "missing"
    send_dir = tmpdir_factory.mktemp("missing_file")
    cfg.cwd = send_dir

    with pytest.raises(TransferError) as e:
        build_offer(cfg)
    assert str(e.value) == f"Cannot send: no file/directory named '{filename}'"


def _do_test_directory(parent_dir, addslash):
    send_dir = "dirname"
    os.mkdir(os.path.join(parent_dir, send_dir))
    ponies = [str(i) for i in range(5)]
    for p in ponies:
        with open(os.path.join(parent_dir, send_dir, p), "wb") as f:
            f.write(f"{p} ponies\n".encode("ascii"))

    send_dir_arg = send_dir
    if addslash:
        send_dir_arg += os.sep
    cfg = create_config()
    cfg.what = send_dir_arg
    cfg.cwd = parent_dir

    d, fd_to_send = build_offer(cfg)

    assert "message" not in d
    assert "file" not in d
    assert "directory" in d
    assert d["directory"]["dirname"] == send_dir
    assert d["directory"]["mode"] == "zipfile/deflated"
    assert d["directory"]["numfiles"] == 5
    assert "numbytes" in d["directory"]
    assert isinstance(d["directory"]["numbytes"], int)

    zdata = b"".join(fd_to_send)
    assert len(zdata) == d["directory"]["zipsize"]
    with zipfile.ZipFile(io.BytesIO(zdata), "r") as zf:
        zipnames = zf.namelist()
        assert list(sorted(ponies)) == list(sorted(zipnames))
        for name in zipnames:
            contents = zf.open(name, "r").read()
            assert (f"{name} ponies\n").encode("ascii") == \
                             contents


def test_directory_simple(tmpdir_factory):
    return _do_test_directory(tmpdir_factory.mktemp("dir"), addslash=False)


def test_directory_addslash_simple(tmpdir_factory):
    return _do_test_directory(tmpdir_factory.mktemp("addslash"), addslash=True)


def test_unknown(request, tmpdir_factory):
    cfg = create_config()
    cfg.what = filename = "unknown"
    send_dir = tmpdir_factory.mktemp("unknown")
    abs_filename = os.path.abspath(os.path.join(send_dir, filename))
    cfg.cwd = send_dir

    try:
        os.mkfifo(abs_filename)
    except AttributeError:
        return pytest.skip("is mkfifo supported on this platform?")

    # Delete the named pipe for the sake of users who might run "pip
    # wheel ." in this directory later. That command wants to copy
    # everything into a tempdir before building a wheel, and the
    # shutil.copy_tree() is uses can't handle the named pipe.
    request.addfinalizer(partial(maybe_delete, abs_filename))

    assert not os.path.isfile(abs_filename)
    assert not os.path.isdir(abs_filename)

    with pytest.raises(TypeError) as e:
        build_offer(cfg)
    assert str(e.value) == f"'{filename}' is neither file nor directory"


@pytest.mark.skipif(not hasattr(os, 'symlink'), reason="host OS does not support symlinks")
def test_symlink(tmpdir_factory):
    # build A/B1 -> B2 (==A/B2), and A/B2/C.txt
    parent_dir = tmpdir_factory.mktemp("symlink_parent")
    os.mkdir(os.path.join(parent_dir, "B2"))
    with open(os.path.join(parent_dir, "B2", "C.txt"), "wb") as f:
        f.write(b"success")
    os.symlink("B2", os.path.join(parent_dir, "B1"))
    # now send "B1/C.txt" from A, and it should get the right file
    cfg = create_config()
    cfg.cwd = parent_dir
    cfg.what = os.path.join("B1", "C.txt")
    d, fd_to_send = build_offer(cfg)
    assert d["file"]["filename"] == "C.txt"
    assert fd_to_send.read() == b"success"


# ntpath.py's realpath() is built out of normpath(), and does not
# follow symlinks properly, so this test always fails. "wormhole send
# PATH" on windows will do the wrong thing. See
# https://bugs.python.org/issue9949" for details. I'm making this a
# TODO instead of a SKIP because 1: this causes an observable
# misbehavior (albeit in rare circumstances), 2: it probably used to
# work (sometimes, but not in #251). See cmd_send.py for more notes.
@pytest.mark.skipif(os.name == "nt", reason="host OS has broken os.path.realpath()")
@pytest.mark.skipif(not hasattr(os, 'symlink'), reason="host OS does not support symlinks")
def test_symlink_collapse(tmpdir_factory):
    cfg = create_config()
    # build A/B1, A/B1/D.txt
    # A/B2/C2, A/B2/D.txt
    # symlink A/B1/C1 -> A/B2/C2
    parent_dir = tmpdir_factory.mktemp("parent")
    os.mkdir(os.path.join(parent_dir, "B1"))
    with open(os.path.join(parent_dir, "B1", "D.txt"), "wb") as f:
        f.write(b"fail")
    os.mkdir(os.path.join(parent_dir, "B2"))
    os.mkdir(os.path.join(parent_dir, "B2", "C2"))
    with open(os.path.join(parent_dir, "B2", "D.txt"), "wb") as f:
        f.write(b"success")
    os.symlink(
        os.path.abspath(os.path.join(parent_dir, "B2", "C2")),
        os.path.join(parent_dir, "B1", "C1"))
    # Now send "B1/C1/../D.txt" from A. The correct traversal will be:
    # * start: A
    # * B1: A/B1
    # * C1: follow symlink to A/B2/C2
    # * ..: climb to A/B2
    # * D.txt: open A/B2/D.txt, which contains "success"
    # If the code mistakenly uses normpath(), it would do:
    # * normpath turns B1/C1/../D.txt into B1/D.txt
    # * start: A
    # * B1: A/B1
    # * D.txt: open A/B1/D.txt , which contains "fail"
    cfg.cwd = parent_dir
    cfg.what = os.path.join("B1", "C1", os.pardir, "D.txt")
    d, fd_to_send = build_offer(cfg)
    assert d["file"]["filename"] == "D.txt"
    assert fd_to_send.read() == b"success"


async def _find_utf8_locale():
    """
    Click really wants to be running under a unicode-capable locale,
    especially on python3. macOS has en-US.UTF-8 but not C.UTF-8, and
    most linux boxes have C.UTF-8 but not en-US.UTF-8 . For tests,
    figure out which one is present and use that. For runtime, it's a
    mess, as really the user must take responsibility for setting their
    locale properly. I'm thinking of abandoning Click and going back to
    twisted.python.usage to avoid this problem in the future.
    """
    (out, err, rc) = await getProcessOutputAndValue("locale", ["-a"])
    if rc != 0:
        log.msg(f"error running 'locale -a', rc={rc}")
        log.msg(f"stderr: {err}")
        return None
    out = out.decode("utf-8")  # make sure we get a string
    utf8_locales = {}
    for locale in out.splitlines():
        locale = locale.strip()
        if locale.lower().endswith((".utf-8", ".utf8")):
            utf8_locales[locale.lower()] = locale
    for wanted in ["C.utf8", "C.UTF-8", "en_US.utf8", "en_US.UTF-8"]:
        if wanted.lower() in utf8_locales:
            return utf8_locales[wanted.lower()]
    if utf8_locales:
        return list(utf8_locales.values())[0]
    return None


@pytest.fixture(scope="module")
def locale():
    if sys.platform == "win32":
        return "en_US.UTF-8"
    best_locale = pytest_twisted.blockon(
        ensureDeferred(_find_utf8_locale())
    )
    yield best_locale


@pytest.fixture(scope="module")
def wormhole_executable():
    """
    to make sure we're running the right executable (in a virtualenv),
    we require that our "wormhole" lives in the same directory as our
    "python"
    """
    locations = procutils.which("wormhole")
    if not locations:
        return pytest.skip("unable to find 'wormhole' in $PATH")
    wormhole = locations[0]
    if (os.path.dirname(os.path.abspath(wormhole)) != os.path.dirname(
            sys.executable)):
        log.msg(f"locations: {locations}")
        log.msg(f"sys.executable: {sys.executable}")
        return pytest.skip(
            f"found the wrong 'wormhole' in $PATH: {wormhole} {sys.executable}")
    return wormhole


@pytest.fixture(scope="module")
def scripts_env(wormhole_executable, locale):
    # One property of Versioneer is that many changes to the source tree
    # (making a commit, dirtying a previously-clean tree) will change the
    # version string. Entrypoint scripts frequently insist upon importing
    # a library version that matches the script version (whatever was
    # reported when 'pip install' was run), and throw a
    # DistributionNotFound error when they don't match. This is really
    # annoying in a workspace created with "pip install -e .", as you
    # must re-run pip after each commit.

    # So let's report just one error in this case (from test_version),
    # and skip the other tests that we know will fail.

    # Setting LANG/LC_ALL to a unicode-capable locale is necessary to
    # convince Click to not complain about a forced-ascii locale. My
    # apologies to folks who want to run tests on a machine that doesn't
    # have the C.UTF-8 locale installed.
    if not locale:
        return pytest.skip("unable to find UTF-8 locale")
    locale_env = dict(LC_ALL=locale, LANG=locale)
    res = pytest_twisted.blockon(
        getProcessOutputAndValue(
            wormhole_executable,
            ["--version"],
            env=locale_env,
        )
    )
    out, err, rc = res
    if rc != 0:
        log.msg("wormhole not runnable in this tree:")
        log.msg("out", out)
        log.msg("err", err)
        log.msg("rc", rc)
        pytest.skip("wormhole is not runnable in this tree")
        return
    return locale_env


@pytest_twisted.ensureDeferred
async def test_version(wormhole_executable):
    """
    "wormhole" must be on the path, so e.g. "pip install -e ." in a
    virtualenv. This guards against an environment where the tests
    below might run the wrong executable.
    """

    # we must pass on the environment so that "something" doesn't
    # get sad about UTF8 vs. ascii encodings
    out, err, rc = await getProcessOutputAndValue(
        wormhole_executable, ["--version"], env=os.environ)
    err = err.decode("utf-8")
    if "DistributionNotFound" in err:
        log.msg(f"stderr was {err}")
        last = err.strip().split("\n")[-1]
        assert False, f"wormhole not runnable: {last}"
    ver = out.decode("utf-8") or err
    assert ver.strip() == f"magic-wormhole {__version__}"
    assert rc == 0


@implementer(ITorManager)
class FakeTor:
    # use normal endpoints, but record the fact that we were asked
    def __init__(self):
        self.endpoints = []

    def stream_via(self, host, port, tls=False):
        self.endpoints.append((host, port, tls))
        return endpoints.HostnameEndpoint(reactor, host, port)


def strip_deprecations(stderr, NL):
    lines = [line
             for line in stderr.split(NL)
             if not ("Python 2 is no longer supported" in line or
                     "from cryptography import utils" in line or
                     "support will be dropped in the next release of cryptography" in line
                     )
             ]
    return NL.join(lines)


# note: this is a straight re-factor from the UnitTest test, but it
# might be better to re-factor further? Can we take "setup" code and
# "teardown" code and provide helpers for those, to get rid of "mode="
# and put the actually-different test code in the tests?


@pytest_twisted.ensureDeferred
async def _do_test(
        wormhole_executable,
        scripts_env,
        mailbox,
        transiturl,
        tmpdir_factory,
        as_subprocess=False,
        mode="text",
        addslash=False,
        override_filename=False,
        fake_tor=False,
        overwrite=False,
        mock_accept=False,
        verify=False):
    assert mode in ("text", "file", "empty-file", "directory", "slow-text",
                    "slow-sender-text")
    if fake_tor:
        assert not as_subprocess
    send_cfg = config("send")
    recv_cfg = config("receive")
    message = "blah blah blah ponies"

    for cfg in [send_cfg, recv_cfg]:
        cfg.hide_progress = True
        cfg.relay_url = mailbox.url
        cfg.transit_helper = ""
        cfg.listen = True
        cfg.code = "1-abc"
        cfg.stdout = io.StringIO()
        cfg.stderr = io.StringIO()
        cfg.verify = verify

    send_dir = tmpdir_factory.mktemp("sender")
    receive_dir = tmpdir_factory.mktemp("receiver")

    if mode in ("text", "slow-text", "slow-sender-text"):
        send_cfg.text = message

    elif mode in ("file", "empty-file"):
        if mode == "empty-file":
            message = ""
        send_filename = "testfil\u00EB"  # e-with-diaeresis
        with open(os.path.join(send_dir, send_filename), "w") as f:
            f.write(message)
        send_cfg.what = send_filename
        receive_filename = send_filename

        recv_cfg.accept_file = False if mock_accept else True
        if override_filename:
            recv_cfg.output_file = receive_filename = "outfile"
        if overwrite:
            recv_cfg.output_file = receive_filename
            existing_file = os.path.join(receive_dir, receive_filename)
            with open(existing_file, 'w') as f:
                f.write('pls overwrite me')

    elif mode == "directory":
        # $send_dir/
        # $send_dir/middle/
        # $send_dir/middle/$dirname/
        # $send_dir/middle/$dirname/[12345]
        # cd $send_dir && wormhole send middle/$dirname
        # cd $receive_dir && wormhole receive
        # expect: $receive_dir/$dirname/[12345]

        send_dirname = "testdir"

        def message(i):
            return "test message %d\n" % i

        os.mkdir(os.path.join(send_dir, "middle"))
        source_dir = os.path.join(send_dir, "middle", send_dirname)
        os.mkdir(source_dir)
        modes = {}
        for i in range(5):
            path = os.path.join(source_dir, str(i))
            with open(path, "w") as f:
                f.write(message(i))
            if i == 3:
                os.chmod(path, 0o755)
            modes[i] = stat.S_IMODE(os.stat(path).st_mode)
        send_dirname_arg = os.path.join("middle", send_dirname)
        if addslash:
            send_dirname_arg += os.sep
        send_cfg.what = send_dirname_arg
        receive_dirname = send_dirname

        recv_cfg.accept_file = False if mock_accept else True
        if override_filename:
            recv_cfg.output_file = receive_dirname = "outdir"
        if overwrite:
            recv_cfg.output_file = receive_dirname
            existing_file = os.path.join(receive_dir, receive_dirname)
            with open(existing_file, 'w') as f:
                f.write('pls overwrite me')

    if as_subprocess:
        if send_cfg.text:
            content_args = ['--text', send_cfg.text]
        elif send_cfg.what:
            content_args = [send_cfg.what]

        # raise the rx KEY_TIMER to some large number here, to avoid
        # spurious test failures on hosts that are slow enough to trigger
        # the "Waiting for sender..." pacifier message. We can do in
        # not-as_subprocess, because we can directly patch the value before
        # running the receiver. But we can't patch across the subprocess
        # boundary, so we use an environment variable.
        env = scripts_env.copy()
        env["_MAGIC_WORMHOLE_TEST_KEY_TIMER"] = "999999"
        env["_MAGIC_WORMHOLE_TEST_VERIFY_TIMER"] = "999999"
        send_args = [
            '--relay-url',
            mailbox.url,
            '--transit-helper',
            '',
            'send',
            '--hide-progress',
            '--code',
            send_cfg.code,
        ] + content_args

        send_d = getProcessOutputAndValue(
            wormhole_executable,
            send_args,
            path=send_dir,
            env=env,
        )
        recv_args = [
            '--relay-url',
            mailbox.url,
            '--transit-helper',
            '',
            'receive',
            '--hide-progress',
            '--accept-file',
            recv_cfg.code,
        ]
        if override_filename:
            recv_args.extend(['-o', receive_filename])

        receive_d = getProcessOutputAndValue(
            wormhole_executable,
            recv_args,
            path=receive_dir,
            env=env,
        )

        (send_res, receive_res) = await gatherResults([send_d, receive_d],
                                                      True)
        send_stdout = send_res[0].decode("utf-8")
        send_stderr = send_res[1].decode("utf-8")
        send_rc = send_res[2]
        receive_stdout = receive_res[0].decode("utf-8")
        receive_stderr = receive_res[1].decode("utf-8")
        receive_rc = receive_res[2]
        NL = os.linesep
        send_stderr = strip_deprecations(send_stderr, NL)
        receive_stderr = strip_deprecations(receive_stderr, NL)
        assert send_rc == 0, f"send failed: {send_res}"
        assert receive_rc == 0, f"receive failed: {receive_res}"

    else:
        send_cfg.cwd = send_dir
        recv_cfg.cwd = receive_dir

        if fake_tor:
            send_cfg.tor = True
            send_cfg.transit_helper = transiturl
            tx_tm = FakeTor()
            with mock.patch(
                    "wormhole.tor_manager.get_tor",
                    return_value=tx_tm,
            ) as mtx_tm:
                send_d = cmd_send.send(send_cfg)

            recv_cfg.tor = True
            recv_cfg.transit_helper = transiturl
            rx_tm = FakeTor()
            with mock.patch(
                    "wormhole.tor_manager.get_tor",
                    return_value=rx_tm,
            ) as mrx_tm:
                receive_d = cmd_receive.receive(recv_cfg)
        else:
            KEY_TIMER = 0 if mode == "slow-sender-text" else 99999
            rxw = []
            with mock.patch.object(cmd_receive, "KEY_TIMER", KEY_TIMER):
                send_d = cmd_send.send(send_cfg)
                receive_d = cmd_receive.receive(recv_cfg, _debug_stash_wormhole=rxw)
                # we need to keep KEY_TIMER patched until the receiver
                # gets far enough to start the timer, which happens after
                # the code is set
                if mode == "slow-sender-text":
                    await rxw[0].get_unverified_key()

        # The sender might fail, leaving the receiver hanging, or vice
        # versa. Make sure we don't wait on one side exclusively
        VERIFY_TIMER = 0 if mode == "slow-text" else 99999
        with mock.patch.object(cmd_receive, "VERIFY_TIMER", VERIFY_TIMER):
            with mock.patch.object(cmd_send, "VERIFY_TIMER", VERIFY_TIMER):
                if mock_accept or verify:
                    with mock.patch.object(builtins, 'input',
                            return_value='yes') as i:
                        await gatherResults([send_d, receive_d], True)
                    if verify:
                        s = i.mock_calls[0][1][0]
                        mo = re.search(r'^Verifier (\w+)\. ok\?', s)
                        assert mo, s
                        sender_verifier = mo.group(1)
                else:
                    await gatherResults([send_d, receive_d], True)

        if fake_tor:
            expected_endpoints = [("127.0.0.1", mailbox.port._realPortNumber, False)]
            if mode in ("file", "directory"):
                transitport = int(transiturl.split(":")[2])
                expected_endpoints.append(("127.0.0.1", transitport, False))
            tx_timing = mtx_tm.call_args[1]["timing"]
            assert tx_tm.endpoints == expected_endpoints
            assert mtx_tm.mock_calls == \
                [mock.call(reactor, False, None, timing=tx_timing)]
            rx_timing = mrx_tm.call_args[1]["timing"]
            assert rx_tm.endpoints == expected_endpoints
            assert mrx_tm.mock_calls == \
                [mock.call(reactor, False, None, timing=rx_timing)]

        send_stdout = send_cfg.stdout.getvalue()
        send_stderr = send_cfg.stderr.getvalue()
        receive_stdout = recv_cfg.stdout.getvalue()
        receive_stderr = recv_cfg.stderr.getvalue()

        # all output here comes from a StringIO, which uses \n for
        # newlines, even if we're on windows
        NL = "\n"

    key_established = ""
    if mode == "slow-text":
        key_established = "Key established, waiting for confirmation...\n"

    assert send_stdout == "Note: code has been consumed and can no longer be used.\n"

    # check sender
    if mode == "text" or mode == "slow-text":
        snippets = [
            "Sending text message ({bytes:d} Bytes){NL}",
            "Wormhole code is: {code}{NL}",
            (
                "On the other computer, please run:{NL}{NL}"
                "wormhole receive {verify}{code}{NL}{NL}"
            ),
            "{KE}",
            "text message sent{NL}",
        ]
        for snippet in snippets:
            expected = snippet.format(
                bytes=len(message),
                verify="--verify " if verify else "",
                code=send_cfg.code,
                NL=NL,
                KE=key_established,
            )
            assert expected in send_stderr
    elif mode == "file":
        expected = "Sending {size:s} file named '{name}'{NL}".format(
            size=naturalsize(len(message)),
            name=send_filename,
            NL=NL)
        assert expected in send_stderr
        assert f"Wormhole code is: {send_cfg.code}{NL}" in send_stderr
        expected = ("On the other computer, please run:{NL}{NL}"
                    "wormhole receive {code}{NL}{NL}").format(code=send_cfg.code, NL=NL)
        assert expected in send_stderr
        assert ("File sent.. waiting for confirmation{NL}"
                "Confirmation received. Transfer complete.{NL}").format(NL=NL) in send_stderr

    elif mode == "directory":
        assert "Sending directory" in send_stderr
        assert "named 'testdir'" in send_stderr
        assert f"Wormhole code is: {send_cfg.code}{NL}" in send_stderr
        assert "On the other computer, please run:{NL}{NL}wormhole receive {code}{NL}{NL}".format(code=send_cfg.code, NL=NL) in send_stderr

        assert "File sent.. waiting for confirmation{NL}Confirmation received. Transfer complete.{NL}".format(NL=NL) in send_stderr

    # check receiver
    if mode in ("text", "slow-text", "slow-sender-text"):
        assert receive_stdout == message + NL
        if mode == "text":
            if verify:
                mo = re.search(r'^Verifier (\w+)\.\s*$', receive_stderr)
                assert mo, receive_stderr
                receiver_verifier = mo.group(1)
                assert sender_verifier == receiver_verifier
            else:
                assert receive_stderr == ""
        elif mode == "slow-text":
            assert receive_stderr == key_established
        elif mode == "slow-sender-text":
            assert receive_stderr == "Waiting for sender...\n"
    elif mode == "file":
        assert receive_stdout == ""
        want = "Receiving file ({size:s}) into: {name!r}".format(
            size=naturalsize(len(message)),
            name=receive_filename,
        )
        assert want in receive_stderr
        assert "Received file written to:" in receive_stderr
        fn = os.path.join(receive_dir, receive_filename)
        assert os.path.exists(fn)
        with open(fn) as f:
            assert f.read() == message
    elif mode == "directory":
        assert receive_stdout == ""
        want = (r"Receiving directory \(\d+ \w+\) into: {name!r}/"
                .format(name=receive_dirname))
        assert re.search(want, receive_stderr), (want, receive_stderr)
        assert "Received files written to" in receive_stderr
        assert f"{receive_dirname!r}" in receive_stderr
        fn = os.path.join(receive_dir, receive_dirname)
        assert os.path.exists(fn), fn
        for i in range(5):
            fn = os.path.join(receive_dir, receive_dirname, str(i))
            with open(fn) as f:
                assert f.read() == message(i)
            assert modes[i] == stat.S_IMODE(
                os.stat(fn).st_mode)

async def test_text(wormhole_executable, scripts_env, mailbox, transit_relay, tmpdir_factory):
    await _do_test(wormhole_executable, scripts_env, mailbox, transit_relay, tmpdir_factory)


async def test_text_subprocess(wormhole_executable, scripts_env, mailbox, transit_relay, tmpdir_factory):
    await _do_test(wormhole_executable, scripts_env, mailbox, transit_relay, tmpdir_factory, as_subprocess=True)


async def test_text_tor(wormhole_executable, scripts_env, mailbox, transit_relay, tmpdir_factory):
    await _do_test(wormhole_executable, scripts_env, mailbox, transit_relay, tmpdir_factory, fake_tor=True)


async def test_text_verify(wormhole_executable, scripts_env, mailbox, transit_relay, tmpdir_factory):
    await _do_test(wormhole_executable, scripts_env, mailbox, transit_relay, tmpdir_factory, verify=True)


async def test_file(wormhole_executable, scripts_env, mailbox, transit_relay, tmpdir_factory):
    await _do_test(wormhole_executable, scripts_env, mailbox, transit_relay, tmpdir_factory, mode="file")


async def test_file_override(wormhole_executable, scripts_env, mailbox, transit_relay, tmpdir_factory):
    await _do_test(wormhole_executable, scripts_env, mailbox, transit_relay, tmpdir_factory, mode="file", override_filename=True)


async def test_file_overwrite(wormhole_executable, scripts_env, mailbox, transit_relay, tmpdir_factory):
    await _do_test(wormhole_executable, scripts_env, mailbox, transit_relay, tmpdir_factory, mode="file", overwrite=True)


async def test_file_overwrite_mock_accept(wormhole_executable, scripts_env, mailbox, transit_relay, tmpdir_factory):
    await _do_test(wormhole_executable, scripts_env, mailbox, transit_relay, tmpdir_factory, mode="file", overwrite=True, mock_accept=True)


async def test_file_tor(wormhole_executable, scripts_env, mailbox, transit_relay, tmpdir_factory):
    await _do_test(wormhole_executable, scripts_env, mailbox, transit_relay, tmpdir_factory, mode="file", fake_tor=True)


async def test_empty_file(wormhole_executable, scripts_env, mailbox, transit_relay, tmpdir_factory):
    await _do_test(wormhole_executable, scripts_env, mailbox, transit_relay, tmpdir_factory, mode="empty-file")


async def test_directory(wormhole_executable, scripts_env, mailbox, transit_relay, tmpdir_factory):
    await _do_test(wormhole_executable, scripts_env, mailbox, transit_relay, tmpdir_factory, mode="directory")


async def test_directory_addslash(wormhole_executable, scripts_env, mailbox, transit_relay, tmpdir_factory):
    await _do_test(wormhole_executable, scripts_env, mailbox, transit_relay, tmpdir_factory, mode="directory", addslash=True)


async def test_directory_override(wormhole_executable, scripts_env, mailbox, transit_relay, tmpdir_factory):
    await _do_test(wormhole_executable, scripts_env, mailbox, transit_relay, tmpdir_factory, mode="directory", override_filename=True)


async def test_directory_overwrite(wormhole_executable, scripts_env, mailbox, transit_relay, tmpdir_factory):
    await _do_test(wormhole_executable, scripts_env, mailbox, transit_relay, tmpdir_factory, mode="directory", overwrite=True)


async def test_slow_text(wormhole_executable, scripts_env, mailbox, transit_relay, tmpdir_factory):
    await _do_test(wormhole_executable, scripts_env, mailbox, transit_relay, tmpdir_factory, mode="slow-text")


async def test_slow_sender_text(wormhole_executable, scripts_env, mailbox, transit_relay, tmpdir_factory):
    await _do_test(wormhole_executable, scripts_env, mailbox, transit_relay, tmpdir_factory, mode="slow-sender-text")


@pytest_twisted.ensureDeferred
async def _do_test_fail(wormhole_executable, scripts_env, relayurl, tmpdir_factory, mode, failmode):
    assert mode in ("file", "directory")
    assert failmode in ("noclobber", "toobig")
    send_cfg = config("send")
    recv_cfg = config("receive")

    for cfg in [send_cfg, recv_cfg]:
        cfg.hide_progress = True
        cfg.relay_url = relayurl
        cfg.transit_helper = ""
        cfg.listen = False
        cfg.code = "1-abc"
        cfg.stdout = io.StringIO()
        cfg.stderr = io.StringIO()

    send_dir = tmpdir_factory.mktemp(f"sendfail-{mode}-{failmode}")
    receive_dir = tmpdir_factory.mktemp(f"recvfail-{mode}-{failmode}")
    recv_cfg.accept_file = True  # don't ask for permission

    if mode == "file":
        message = "test message\n"
        send_cfg.what = receive_name = send_filename = "testfile"
        fn = os.path.join(send_dir, send_filename)
        with open(fn, "w") as f:
            f.write(message)
        size = os.stat(fn).st_size

    elif mode == "directory":
        # $send_dir/
        # $send_dir/$dirname/
        # $send_dir/$dirname/[12345]
        # cd $send_dir && wormhole send $dirname
        # cd $receive_dir && wormhole receive
        # expect: $receive_dir/$dirname/[12345]

        size = 0
        send_cfg.what = receive_name = send_dirname = "testdir"
        os.mkdir(os.path.join(send_dir, send_dirname))
        for i in range(5):
            path = os.path.join(send_dir, send_dirname, str(i))
            with open(path, "w") as f:
                f.write("test message %d\n" % i)
            size += os.stat(path).st_size

    if failmode == "noclobber":
        PRESERVE = "don't clobber me\n"
        clobberable = os.path.join(receive_dir, receive_name)
        with open(clobberable, "w") as f:
            f.write(PRESERVE)

    send_cfg.cwd = send_dir
    send_d = cmd_send.send(send_cfg)

    recv_cfg.cwd = receive_dir
    receive_d = cmd_receive.receive(recv_cfg)

    # both sides will fail
    if failmode == "noclobber":
        free_space = 10000000
    else:
        free_space = 0
    with mock.patch(
            "wormhole.cli.cmd_receive.estimate_free_space",
            return_value=free_space):
        with pytest.raises(TransferError) as e:
            await send_d
        assert str(e.value) == "remote error, transfer abandoned: transfer rejected"

        with pytest.raises(TransferError) as e:
            await receive_d
        assert str(e.value) == "transfer rejected"

    send_stdout = send_cfg.stdout.getvalue()
    send_stderr = send_cfg.stderr.getvalue()
    receive_stdout = recv_cfg.stdout.getvalue()
    receive_stderr = recv_cfg.stderr.getvalue()

    # all output here comes from a StringIO, which uses \n for
    # newlines, even if we're on windows
    NL = "\n"

    assert send_stdout == "Note: code has been consumed and can no longer be used.\n"
    assert receive_stdout == ""

    # check sender
    if mode == "file":
        assert "Sending {size:s} file named '{name}'{NL}".format(
            size=naturalsize(size),
            name=send_filename,
            NL=NL) in send_stderr
        assert f"Wormhole code is: {send_cfg.code}{NL}" in send_stderr
        assert "On the other computer, please run:{NL}{NL}wormhole receive {code}{NL}{NL}".format(code=send_cfg.code, NL=NL) in send_stderr
    elif mode == "directory":
        assert "Sending directory" in send_stderr
        assert "named 'testdir'" in send_stderr
        assert f"Wormhole code is: {send_cfg.code}{NL}" in send_stderr
        assert "On the other computer, please run:{NL}{NL}wormhole receive {code}{NL}{NL}".format(code=send_cfg.code, NL=NL) in send_stderr

    # check receiver
    if mode == "file":
        assert "Received file written to:" not in receive_stderr
        if failmode == "noclobber":
            assert f"Error: refusing to overwrite existing 'testfile'{NL}" in receive_stderr
        else:
            assert f"Error: insufficient free space (0B) for file ({size:d}B){NL}" in receive_stderr
    elif mode == "directory":
        assert "Received files written to" not in receive_stderr
        # want = (r"Receiving directory \(\d+ \w+\) into: {name}/"
        #        .format(name=receive_name))
        # self.failUnless(re.search(want, receive_stderr),
        #                (want, receive_stderr))
        if failmode == "noclobber":
            assert f"Error: refusing to overwrite existing 'testdir'{NL}" in receive_stderr
        else:
            assert (
                "Error: "
                "insufficient free space (0B) for directory"
                " ({size:d}B){NL}").format(NL=NL, size=size) in receive_stderr

    if failmode == "noclobber":
        fn = os.path.join(receive_dir, receive_name)
        assert os.path.exists(fn)
        with open(fn) as f:
            assert f.read() == PRESERVE


async def test_fail_file_noclobber(wormhole_executable, scripts_env, mailbox, tmpdir_factory):
    await _do_test_fail(wormhole_executable, scripts_env, mailbox.url, tmpdir_factory, "file", "noclobber")


async def test_fail_directory_noclobber(wormhole_executable, scripts_env, mailbox, tmpdir_factory):
    await _do_test_fail(wormhole_executable, scripts_env, mailbox.url, tmpdir_factory, "directory", "noclobber")


async def test_fail_file_toobig(wormhole_executable, scripts_env, mailbox, tmpdir_factory):
    await _do_test_fail(wormhole_executable, scripts_env, mailbox.url, tmpdir_factory, "file", "toobig")


async def test_fail_directory_toobig(wormhole_executable, scripts_env, mailbox, tmpdir_factory):
    await _do_test_fail(wormhole_executable, scripts_env, mailbox.url, tmpdir_factory, "directory", "toobig")


@pytest_twisted.ensureDeferred
async def test_text_ponies(mailbox):
    send_cfg = config("send")
    recv_cfg = config("receive")
    message = "textponies"

    for cfg in [send_cfg, recv_cfg]:
        cfg.hide_progress = True
        cfg.relay_url = mailbox.url
        cfg.transit_helper = ""
        cfg.listen = True
        cfg.zeromode = True
        cfg.stdout = io.StringIO()
        cfg.stderr = io.StringIO()

    send_cfg.text = message

    # send_cfg.cwd = send_dir
    # recv_cfg.cwd = receive_dir

    send_d = cmd_send.send(send_cfg)
    receive_d = cmd_receive.receive(recv_cfg)

    await gatherResults([send_d, receive_d], True)

    send_stdout = send_cfg.stdout.getvalue()
    send_stderr = send_cfg.stderr.getvalue()
    receive_stdout = recv_cfg.stdout.getvalue()
    receive_stderr = recv_cfg.stderr.getvalue()

    # all output here comes from a StringIO, which uses \n for
    # newlines, even if we're on windows
    NL = "\n"

    assert send_stdout == "Note: code has been consumed and can no longer be used.\n"

    # check sender
    expected = ("Sending text message ({bytes:d} Bytes){NL}"
                "On the other computer, please run:{NL}"
                "{NL}"
                "wormhole receive -0{NL}"
                "{NL}"
                "text message sent{NL}").format(
                    bytes=len(message), NL=NL)
    assert send_stderr == expected
    # check receiver
    assert receive_stdout == message + NL
    assert receive_stderr == ""


@pytest.fixture(scope="module")
def unwelcome_mailbox(reactor):
    mailbox = pytest_twisted.blockon(
        ensureDeferred(
            setup_mailbox(reactor, error="please upgrade XYZ")
        )
    )
    mailbox.service.startService()
    yield mailbox.url
    pytest_twisted.blockon(mailbox.service.stopService())


@pytest.fixture()
def unwelcome_config(unwelcome_mailbox):
    cfg = config("send")
    cfg.hide_progress = True
    cfg.listen = False
    cfg.relay_url = unwelcome_mailbox
    cfg.transit_helper = ""
    cfg.stdout = io.StringIO()
    cfg.stderr = io.StringIO()
    return cfg


@pytest_twisted.ensureDeferred
async def test_sender_unwelcome(unwelcome_config):
    unwelcome_config.text = "hi"
    unwelcome_config.code = "1-abc"
    unwelcome_config.verbose = True
    unwelcome_config.stdout = sys.stdout
    unwelcome_config.stderr = sys.stderr

    send_d = cmd_send.send(unwelcome_config)
    with pytest.raises(WelcomeError) as e:
        await send_d
    assert str(e.value) == "please upgrade XYZ"


@pytest_twisted.ensureDeferred
async def test_receiver_unwelcome(unwelcome_config):
    unwelcome_config.code = "1-abc"

    receive_d = cmd_receive.receive(unwelcome_config)
    with pytest.raises(WelcomeError) as e:
        await receive_d
    assert str(e.value) == "please upgrade XYZ"


@pytest.fixture(scope="module")
def no_mailbox(reactor):
    # original test created and then destroyed a mailbox server --
    # basically just to get a port number. Having a hard time
    # "creating and killing" correctly in here, so short-cutting to "a
    # port that shouldn't be listening"
    yield "ws://127.0.0.1:1/v1"


@pytest_twisted.ensureDeferred
async def test_sender(no_mailbox):
    cfg = config("send")
    cfg.hide_progress = True
    cfg.listen = False
    cfg.relay_url = no_mailbox
    cfg.transit_helper = ""
    cfg.stdout = io.StringIO()
    cfg.stderr = io.StringIO()

    cfg.text = "hi"
    cfg.code = "1-abc"

    with pytest.raises(ServerConnectionError) as e:
        await cmd_send.send(cfg)
    assert isinstance(e.value.reason, ConnectionRefusedError)


@pytest_twisted.ensureDeferred
async def test_sender_allocation(no_mailbox):
    cfg = config("send")
    cfg.hide_progress = True
    cfg.listen = False
    cfg.relay_url = no_mailbox
    cfg.transit_helper = ""
    cfg.stdout = io.StringIO()
    cfg.stderr = io.StringIO()

    cfg.text = "hi"

    with pytest.raises(ServerConnectionError) as e:
        await cmd_send.send(cfg)
    assert isinstance(e.value.reason, ConnectionRefusedError)


@pytest_twisted.ensureDeferred
async def test_receiver(no_mailbox):
    cfg = config("receive")
    cfg.hide_progress = True
    cfg.listen = False
    cfg.relay_url = no_mailbox
    cfg.transit_helper = ""
    cfg.stdout = io.StringIO()
    cfg.stderr = io.StringIO()

    cfg.code = "1-abc"

    with pytest.raises(ServerConnectionError) as e:
        await cmd_receive.receive(cfg)
    assert isinstance(e.value.reason, ConnectionRefusedError)


def create_named_config(name, url):
    cfg = config(name)
    # common options for all tests in this suite
    cfg.hide_progress = True
    cfg.relay_url = url
    cfg.transit_helper = ""
    cfg.stdout = io.StringIO()
    cfg.stderr = io.StringIO()
    return cfg


@pytest_twisted.ensureDeferred
async def test_text_send(mailbox):
    send_config = create_named_config("send", mailbox.url)
    recv_config = create_named_config("receive", mailbox.url)
    send_config.code = "1-test-situation"
    recv_config.code = "1-test-situation"
    with mock.patch('sys.stdout'):
        # the rendezvous channel should be deleted after success
        send_config.text = "some text to send"
        send_d = cmd_send.send(send_config)
        receive_d = cmd_receive.receive(recv_config)

        await send_d
        await receive_d

        cids = mailbox.rendezvous.get_app(cmd_send.APPID).get_nameplate_ids()
        assert len(cids) == 0


@pytest_twisted.ensureDeferred
async def test_text_wrong_password(mailbox):
    # if the password was wrong, the rendezvous channel should still be
    # deleted
    send_config = create_named_config("send", mailbox.url)
    send_config.code = "1-foo-bar"
    send_config.text = "some text to send"
    send_d = cmd_send.send(send_config)

    rx_cfg = create_named_config("send", mailbox.url)
    rx_cfg.code = "1-WRONG"
    receive_d = cmd_receive.receive(rx_cfg)

    # both sides should be capable of detecting the mismatch
    with pytest.raises(WrongPasswordError):
        await send_d
    with pytest.raises(WrongPasswordError):
        await receive_d

    cids = mailbox.rendezvous.get_app(cmd_send.APPID).get_nameplate_ids()
    assert len(cids) == 0


def test_filenames(tmpdir_factory):
    args = mock.Mock()
    args.relay_url = ""
    args.output_file = None
    ef = cmd_receive.Receiver(args)._extract_file
    extract_dir = os.path.abspath(tmpdir_factory.mktemp("filenames"))
    extract_base = os.path.basename(extract_dir)

    zf = mock.Mock()
    zi = mock.Mock()
    zi.filename = "ok"
    zi.external_attr = 5 << 16
    expected = os.path.join(extract_dir, "ok")
    with mock.patch.object(cmd_receive.os, "chmod") as chmod:
        ef(zf, zi, extract_dir)
        assert zf.extract.mock_calls == [mock.call(zi.filename, path=extract_dir)]
        assert chmod.mock_calls == [mock.call(expected, 5)]

    # these test that the receiver prevents "Zip Slip", where zipfile
    # members contain ".." or path-resetting extra slashes, to escape
    # the target directory. The zipfile module already strips
    # leading-slash, double-slash, ".", and "..", making the output safe
    # against those, but we go further and throw an error if we see them

    # basic reach-for-root attack
    zf = mock.Mock()
    zi = mock.Mock()
    zi.filename = "/etc/passwd"
    with pytest.raises(ValueError) as e:
        ef(zf, zi, extract_dir)
    assert "malicious zipfile" in str(e.value)

    # basic climb-the-tree attack
    zf = mock.Mock()
    zi = mock.Mock()
    zi.filename = "../haha"
    with pytest.raises(ValueError) as e:
        ef(zf, zi, extract_dir)
        assert "malicious zipfile" in str(e)

    # sneaky form that reaches for a sibling of the starting directory
    # whose name is an extension of our own, like using
    # "../target2/foo.txt" to escape a "target" directory. If the check
    # in cmd_receive.py uses .startswith() *without* an extra / or
    # os.sep, it won't catch this, and zipfile won't either.
    zf = mock.Mock()
    zi = mock.Mock()
    zi.filename = "../{}-plus-hyphen/haha".format(extract_base)
    zi.external_attr = 0o600 << 16
    with mock.patch.object(cmd_receive.os, "chmod") as chmod:
        with pytest.raises(ValueError) as e:
            ef(zf, zi, extract_dir)
        assert "malicious zipfile" in str(e)

    # this form uses an internal extra slash to attempt access to the
    # filesystem root, we rely on zipfile's handling (no error)
    zf = mock.Mock()
    zi = mock.Mock()
    # abspath squashes this, hopefully zipfile does too
    zi.filename = "haha//root"
    zi.external_attr = 5 << 16
    expected = os.path.join(extract_dir, "haha", "root")
    with mock.patch("os.chmod") as chmod:
        ef(zf, zi, extract_dir)
        assert zf.extract.mock_calls, [mock.call(zi.filename, path=extract_dir)]
        assert chmod.mock_calls == [mock.call(expected, 5)]


    expected = os.path.join(extract_dir, "evil.txt")
    receiver = cmd_receive.Receiver(args)
    receiver.args.cwd = extract_dir
    result = receiver._decide_destname("file", "../../evil.txt")  # basename strips path traversal
    assert expected == result


def test_existing_destdir(tmpdir_factory):
    """
    We should preserve user data when they specify an existing
    destination _directory_ via --output-file (whereas we overwrite
    files if explicitly specified like this)
    """
    args = mock.Mock()
    args.relay_url = ""
    tmpdir = tempfile.mkdtemp()
    args.cwd = os.getcwd()
    args.output_file = tmpdir
    cmd = cmd_receive.Receiver(args)

    s = cmd._decide_destname(None, "destination_file")
    assert s == os.path.join(tmpdir, "destination_file")


def test_existing_destdir_malicious(tmpdir_factory):
    """
    Test that our attempts to preserve an existing output
    directory correctly work even if a zip contains a (possibly)
    malicious path pointing outside of the destdir (e.g. via ../ and
    similar)
    """
    args = mock.Mock()
    args.relay_url = ""
    tmpdir = tempfile.mkdtemp()
    args.cwd = os.getcwd()
    args.output_file = tmpdir
    cmd = cmd_receive.Receiver(args)

    s = cmd._decide_destname(None, "../../../destination_file")
    assert s == os.path.join(tmpdir, "destination_file")


@given(
    st.lists(
        st.one_of(
            st.just(".."),
            st.text(),
        )
    ),
    st.one_of([st.just("file"), st.just("directory")]),
    st.booleans(),
)
def test_destdir_traversal(tmpdir_factory, segments, mode, outdir):
    """
    Let Hypothesis explore the space of directories, and ensure we
    can't traverse outside of our base.

    segments: random path segments, including ".."'s
    mode: "file" or "directory" (currently unused by _decide_destname)
    outdir: whether to include an arg for pre-existing output directory
    """
    base = tmpdir_factory.mktemp("base")
    args = mock.Mock()
    args.relay_url = ""
    if outdir:
        args.output_file = base
    else:
        args.output_file = None

    receiver = cmd_receive.Receiver(args)
    receiver.args.cwd = base

    try:
        result = receiver._decide_destname(mode, "/".join(segments))
    except cmd_receive.TransferRejectedError:
        # in this case, we wouldn't even have tried to write to the
        # path, so whatever happened to get here (path-wise) is fine
        return
    rpath = Path(result)
    bpath = Path(base)
    assert rpath.is_relative_to(bpath), "_decide_destname must always decide a subpath"


def test_not_remove_existing_destdir(tmpdir_factory):
    """
    Do not remove an entire existing directory.
    """
    args = mock.Mock()
    args.relay_url = ""
    tmpdir = tempfile.mkdtemp()
    args.cwd = os.getcwd()
    args.output_file = tmpdir
    cmd = cmd_receive.Receiver(args)
    with pytest.raises(cmd_receive.TransferRejectedError):
        cmd._remove_existing(tmpdir)


@pytest_twisted.ensureDeferred
async def test_override(request, reactor):
    # note: we do not use the "mailbox" fixture because that is
    # session-wide and we want to ensure that _JUST_ "appid2"
    # nameplates appear
    mailbox = await setup_mailbox(reactor)
    mailbox.service.startService()

    def cleanup():
        pytest_twisted.blockon(mailbox.service.stopService())
    request.addfinalizer(cleanup)

    cfg = create_named_config("send", mailbox.url)
    cfg.text = "hello"
    cfg.appid = "appid2"
    cfg.code = "1-abc"
    send_d = ensureDeferred(cmd_send.send(cfg))
    receive_d = ensureDeferred(cmd_receive.receive(cfg))

    await gatherResults([send_d, receive_d])

    used = mailbox.usage_db.execute(
        "SELECT DISTINCT `app_id` FROM `nameplates`").fetchall()
    assert len(used) == 1, f"Incorrect nameplates: {used}"
    assert used[0]["app_id"] == "appid2"


def _welcome_test(welcome_message, my_version="2.0"):
    stderr = io.StringIO()
    welcome.handle_welcome(welcome_message, "url", my_version, stderr)
    return stderr.getvalue()


def test_empty():
    stderr = _welcome_test({})
    assert stderr == ""


def test_version_current():
    stderr = _welcome_test({"current_cli_version": "2.0"})
    assert stderr == ""


def test_version_old():
    stderr = _welcome_test({"current_cli_version": "3.0"})
    expected = ("Warning: errors may occur unless both sides are"
                " running the same version\n"
                "Server claims 3.0 is current, but ours is 2.0\n")
    assert stderr == expected


def test_version_unreleased():
    stderr = _welcome_test(
        {
            "current_cli_version": "3.0"
        }, my_version="2.5+middle.something")
    assert stderr == ""


def test_motd():
    stderr = _welcome_test({"motd": "hello"})
    assert stderr == "Server (at url) says:\n hello\n"


@pytest_twisted.ensureDeferred
async def test_success(mailbox):
    cfg = create_named_config("send", mailbox.url)
    cfg.stderr = io.StringIO()
    called = []

    def fake():
        called.append(1)

    await cli._dispatch_command(reactor, cfg, fake)
    assert called == [1]
    assert cfg.stderr.getvalue() == ""


@pytest_twisted.ensureDeferred
async def test_timing(mailbox):
    cfg = create_named_config("send", mailbox.url)
    cfg.stderr = io.StringIO()
    cfg.timing = mock.Mock()
    cfg.dump_timing = "filename"

    def fake():
        pass

    await cli._dispatch_command(reactor, cfg, fake)
    assert cfg.stderr.getvalue() == ""
    assert cfg.timing.mock_calls[-1] == \
                     mock.call.write("filename", cfg.stderr)


def test_debug_state_invalid_machine():
    cfg = cli.Config()
    with pytest.raises(UsageError):
        cfg.debug_state = "ZZZ"


@pytest_twisted.ensureDeferred
async def test_debug_state_send(mailbox):
    args = create_named_config("send", mailbox.url)
    args.debug_state = "B,N,M,S,O,K,SK,R,RC,L,C,T"
    args.stdout = io.StringIO()
    s = cmd_send.Sender(args, reactor)
    d = s.go()
    d.cancel()
    try:
        await d
    except CancelledError:
        pass
    # just check for at least one state-transition we expected to
    # get logged due to the --debug-state option
    assert "send.B[S0_empty].close" in \
        args.stdout.getvalue()


@pytest_twisted.ensureDeferred
async def test_debug_state_receive(mailbox):
    args = create_named_config("receive", mailbox.url)
    args.debug_state = "B,N,M,S,O,K,SK,R,RC,L,C,T"
    args.stdout = io.StringIO()
    s = cmd_receive.Receiver(args, reactor)
    d = s.go()
    d.cancel()
    try:
        await d
    except CancelledError:
        pass
    # just check for at least one state-transition we expected to
    # get logged due to the --debug-state option
    assert "recv.B[S0_empty].close" in \
        args.stdout.getvalue()


@pytest_twisted.ensureDeferred
async def test_wrong_password_error(mailbox):
    cfg = create_named_config("send", mailbox.url)
    cfg.stderr = io.StringIO()

    def fake():
        raise WrongPasswordError("abcd")

    with pytest.raises(SystemExit):
        await cli._dispatch_command(reactor, cfg, fake)
    expected = fill("ERROR: " + dedent(WrongPasswordError.__doc__)) + "\n"
    assert cfg.stderr.getvalue() == expected


@pytest_twisted.ensureDeferred
async def test_welcome_error(mailbox):
    cfg = create_named_config("send", mailbox.url)
    cfg.stderr = io.StringIO()

    def fake():
        raise WelcomeError("abcd")

    with pytest.raises(SystemExit):
        await cli._dispatch_command(reactor, cfg, fake)
    expected = (
        fill("ERROR: " + dedent(WelcomeError.__doc__)) + "\n\nabcd\n")
    assert cfg.stderr.getvalue() == expected


@pytest_twisted.ensureDeferred
async def test_transfer_error(mailbox):
    cfg = create_named_config("send", mailbox.url)
    cfg.stderr = io.StringIO()

    def fake():
        raise TransferError("abcd")

    with pytest.raises(SystemExit):
        await cli._dispatch_command(reactor, cfg, fake)
    expected = "TransferError: abcd\n"
    assert cfg.stderr.getvalue() == expected


@pytest_twisted.ensureDeferred
async def test_server_connection_error(mailbox):
    cfg = create_named_config("send", mailbox.url)
    cfg.stderr = io.StringIO()

    def fake():
        raise ServerConnectionError("URL", ValueError("abcd"))

    with pytest.raises(SystemExit):
        await cli._dispatch_command(reactor, cfg, fake)
    expected = fill(
        "ERROR: " + dedent(ServerConnectionError.__doc__)) + "\n"
    expected += "(relay URL was URL)\n"
    expected += "abcd\n"
    assert cfg.stderr.getvalue() == expected


@pytest_twisted.ensureDeferred
async def test_other_error(mailbox):
    cfg = create_named_config("send", mailbox.url)
    cfg.stderr = io.StringIO()

    def fake():
        raise ValueError("abcd")

    # I'm seeing unicode problems with the Failure().printTraceback, and
    # the output would be kind of unpredictable anyways, so we'll mock it
    # out here.
    f = mock.Mock()

    def mock_print(file):
        file.write("<TRACEBACK>\n")

    f.printTraceback = mock_print
    with mock.patch("wormhole.cli.cli.Failure", return_value=f):
        with pytest.raises(SystemExit):
            await cli._dispatch_command(reactor, cfg, fake)
    expected = "<TRACEBACK>\nERROR: abcd\n"
    assert cfg.stderr.getvalue() == expected


def _check_top_level_help(got):
    # the main wormhole.cli.cli.wormhole docstring should be in the
    # output, but formatted differently
    assert "Create a Magic Wormhole and communicate through it." in got
    assert "--relay-url" in got
    assert "Receive a text message, file, or directory" in got


def test_help():
    result = CliRunner().invoke(cli.wormhole, ["help"])
    _check_top_level_help(result.output)
    assert result.exit_code == 0


def test_dash_dash_help():
    result = CliRunner().invoke(cli.wormhole, ["--help"])
    _check_top_level_help(result.output)
    assert result.exit_code == 0


def test_inconsistent_receive_code_length():
    """
    specifying --code-length without --allocate is an error
    """
    result = CliRunner().invoke(
        cli.wormhole,
        ["receive", "--code-length", "3", "2-foo-bar"]
    )
    assert result.exit_code != 0
    assert "Must use --allocate" in result.output


def test_inconsistent_receive_allocate():
    """
    specifying --allocate and a code is an error
    """
    result = CliRunner().invoke(
        cli.wormhole,
        ["receive", "--allocate", "2-foo-bar"]
    )
    assert result.exit_code != 0
    assert "Cannot specify a code" in result.output



# no unicode_literals
# Find all of our ip addresses. From tahoe's src/allmydata/util/iputil.py

import errno
import os
import re
import subprocess
from sys import platform

from twisted.python.procutils import which

# Wow, I'm really amazed at home much mileage we've gotten out of calling
# the external route.exe program on windows...  It appears to work on all
# versions so far.  Still, the real system calls would much be preferred...
# ... thus wrote Greg Smith in time immemorial...
_win32_re = re.compile(
    (r'^\s*\d+\.\d+\.\d+\.\d+\s.+\s'
     r'(?P<address>\d+\.\d+\.\d+\.\d+)\s+(?P<metric>\d+)\s*$'),
    flags=re.M | re.I | re.S)
_win32_commands = (('route.exe', ('print', ), _win32_re), )

# These work in most Unices.
_addr_re = re.compile(
    r'^\s*inet [a-zA-Z]*:?(?P<address>\d+\.\d+\.\d+\.\d+)[\s/].+$',
    flags=re.M | re.I | re.S)
_unix_commands = (
    ('/bin/ip', ('addr', ), _addr_re),
    ('/sbin/ip', ('addr', ), _addr_re),
    ('/sbin/ifconfig', ('-a', ), _addr_re),
    ('/usr/sbin/ifconfig', ('-a', ), _addr_re),
    ('/usr/etc/ifconfig', ('-a', ), _addr_re),
    ('ifconfig', ('-a', ), _addr_re),
    ('/sbin/ifconfig', (), _addr_re),
)


def find_addresses():
    # originally by Greg Smith, hacked by Zooko and then Daira

    # We don't reach here for cygwin.
    if platform == 'win32':
        commands = _win32_commands
    else:
        commands = _unix_commands

    for (pathtotool, args, regex) in commands:
        # If pathtotool is a fully qualified path then we just try that.
        # If it is merely an executable name then we use Twisted's
        # "which()" utility and try each executable in turn until one
        # gives us something that resembles a dotted-quad IPv4 address.

        if os.path.isabs(pathtotool):
            exes_to_try = [pathtotool]
        else:
            exes_to_try = which(pathtotool)

        for exe in exes_to_try:
            try:
                addresses = _query(exe, args, regex)
            except Exception:
                addresses = []
            if addresses:
                return addresses

    return ["127.0.0.1"]


def _query(path, args, regex):
    env = {'LANG': 'en_US.UTF-8'}
    trial = 0
    while True:
        trial += 1
        try:
            p = subprocess.Popen(
                [path] + list(args),
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                env=env,
                universal_newlines=True)
            (output, err) = p.communicate()
            break
        except OSError as e:
            if e.errno == errno.EINTR and trial < 5:
                continue
            raise

    addresses = []
    outputsplit = output.split('\n')
    for outline in outputsplit:
        m = regex.match(outline)
        if m:
            addr = m.group('address')
            if addr not in addresses:
                addresses.append(addr)

    return addresses



from attr import attrib, attrs
from wormhole.util import provides
from automat import MethodicalMachine
from zope.interface import implementer

from . import _interfaces


@attrs
@implementer(_interfaces.IAllocator)
class Allocator:
    _timing = attrib(validator=provides(_interfaces.ITiming))
    m = MethodicalMachine()
    set_trace = getattr(m, "_setTrace",
                        lambda self, f: None)  # pragma: no cover

    def wire(self, rendezvous_connector, code):
        self._RC = _interfaces.IRendezvousConnector(rendezvous_connector)
        self._C = _interfaces.ICode(code)

    @m.state(initial=True)
    def S0A_idle(self):
        pass  # pragma: no cover

    @m.state()
    def S0B_idle_connected(self):
        pass  # pragma: no cover

    @m.state()
    def S1A_allocating(self):
        pass  # pragma: no cover

    @m.state()
    def S1B_allocating_connected(self):
        pass  # pragma: no cover

    @m.state()
    def S2_done(self):
        pass  # pragma: no cover

    # from Code
    @m.input()
    def allocate(self, length, wordlist):
        pass

    # from RendezvousConnector
    @m.input()
    def connected(self):
        pass

    @m.input()
    def lost(self):
        pass

    @m.input()
    def rx_allocated(self, nameplate):
        pass

    @m.output()
    def stash(self, length, wordlist):
        self._length = length
        self._wordlist = _interfaces.IWordlist(wordlist)

    @m.output()
    def stash_and_RC_rx_allocate(self, length, wordlist):
        self._length = length
        self._wordlist = _interfaces.IWordlist(wordlist)
        self._RC.tx_allocate()

    @m.output()
    def RC_tx_allocate(self):
        self._RC.tx_allocate()

    @m.output()
    def build_and_notify(self, nameplate):
        words = self._wordlist.choose_words(self._length)
        code = nameplate + "-" + words
        self._C.allocated(nameplate, code)

    S0A_idle.upon(connected, enter=S0B_idle_connected, outputs=[])
    S0B_idle_connected.upon(lost, enter=S0A_idle, outputs=[])

    S0A_idle.upon(allocate, enter=S1A_allocating, outputs=[stash])
    S0B_idle_connected.upon(
        allocate,
        enter=S1B_allocating_connected,
        outputs=[stash_and_RC_rx_allocate])

    S1A_allocating.upon(
        connected, enter=S1B_allocating_connected, outputs=[RC_tx_allocate])
    S1B_allocating_connected.upon(lost, enter=S1A_allocating, outputs=[])

    S1B_allocating_connected.upon(
        rx_allocated, enter=S2_done, outputs=[build_and_notify])

    S2_done.upon(connected, enter=S2_done, outputs=[])
    S2_done.upon(lost, enter=S2_done, outputs=[])



import zope.interface
from attr import Attribute
from attr._make import NOTHING
from wormhole.util import provides

import pytest

class IFoo(zope.interface.Interface):
    """
    An interface.
    """

    def f():
        """
        A function called f.
        """


def simple_attr(name):
    return Attribute(
        name=name,
        default=NOTHING,
        validator=None,
        repr=True,
        cmp=None,
        eq=True,
        hash=None,
        init=True,
        converter=None,
        kw_only=False,
        inherited=False,
    )


def test_success():
    """
    Nothing happens if value provides requested interface.
    """

    @zope.interface.implementer(IFoo)
    class C:
        def f(self):
            pass

    v = provides(IFoo)
    v(None, simple_attr("x"), C())

def test_fail():
    """
    Raises `TypeError` if interfaces isn't provided by value.
    """
    value = object()
    a = simple_attr("x")

    v = provides(IFoo)
    with pytest.raises(TypeError):
        v(None, a, value)

def test_repr():
    """
    Returned validator has a useful `__repr__`.
    """
    v = provides(IFoo)
    assert (
        f"<provides validator for interface {IFoo!r}>"
    ) == repr(v)



from collections import deque, defaultdict
from attr import attrs, attrib
from attr.validators import instance_of
from zope.interface import implementer
from twisted.internet.defer import inlineCallbacks
from twisted.internet.interfaces import (ITransport, IProducer, IConsumer,
                                         IAddress, IListeningPort,
                                         IHalfCloseableProtocol,
                                         IStreamClientEndpoint,
                                         IStreamServerEndpoint,
                                         )
from twisted.internet.error import ConnectionDone
from automat import MethodicalMachine
from .._interfaces import ISubChannel, IDilationManager
from ..util import provides

# each subchannel frame (the data passed into transport.write(data)) gets a
# 9-byte header prefix (type, subchannel id, and sequence number), then gets
# encrypted (adding a 16-byte authentication tag). The result is transmitted
# with a 4-byte length prefix (which only covers the padded message, not the
# length prefix itself), so the padded message must be less than 2**32 bytes
# long.
MAX_FRAME_LENGTH = 2**32 - 1 - 9 - 16


# created in the (OPEN) state, by either:
#  * receipt of an OPEN message
#  * or local client_endpoint.connect()
# then transitions are:
# (OPEN) rx DATA: deliver .dataReceived(), -> (OPEN)
# (OPEN) rx CLOSE: deliver .connectionLost(), send CLOSE, -> (CLOSED)
# (OPEN) local .write(): send DATA, -> (OPEN)
# (OPEN) local .loseConnection(): send CLOSE, -> (CLOSING)
# (CLOSING) local .write(): error
# (CLOSING) local .loseConnection(): error
# (CLOSING) rx DATA: deliver .dataReceived(), -> (CLOSING)
# (CLOSING) rx CLOSE: deliver .connectionLost(), -> (CLOSED)
# object is deleted upon transition to (CLOSED)


class AlreadyClosedError(Exception):
    pass


class NormalCloseUsedOnHalfCloseable(Exception):
    pass


class HalfCloseUsedOnNonHalfCloseable(Exception):
    pass


class UnexpectedSubprotocol(Exception):
    """
    The peer sends an OPEN for a subprotocol name application code
    indicates will never exist
    """


@implementer(IAddress)
class _WormholeAddress:
    pass


@implementer(IAddress)
@attrs
class SubchannelAddress:
    """
    All subchannels have a named sub-protocol (as sent by our peer in
    the OPEN call).

    Although each subchannel does have an 'id', this is a concept
    private to this implementation, and not exposed here on purpose.
    """
    subprotocol = attrib(validator=instance_of(str))


@attrs(eq=False)
@implementer(ITransport)
@implementer(IProducer)
@implementer(IConsumer)
@implementer(ISubChannel)
class SubChannel:
    _scid = attrib(validator=instance_of(int))
    _manager = attrib(validator=provides(IDilationManager))
    _host_addr = attrib(validator=instance_of(_WormholeAddress))
    _peer_addr = attrib(validator=instance_of(SubchannelAddress))

    m = MethodicalMachine()
    set_trace = getattr(m, "_setTrace", lambda self,
                        f: None)  # pragma: no cover

    def __attrs_post_init__(self):
        # self._mailbox = None
        # self._pending_outbound = {}
        # self._processed = set()
        self._protocol = None
        self._pending_remote_data = []
        self._pending_remote_close = False

    @m.state(initial=True)
    def unconnected(self):
        pass  # pragma: no cover

    # once we get the IProtocol, it's either a IHalfCloseableProtocol, or it
    # can only be fully closed
    @m.state()
    def open_half(self):
        pass  # pragma: no cover

    @m.state()
    def read_closed():
        pass  # pragma: no cover

    @m.state()
    def write_closed():
        pass  # pragma: no cover

    @m.state()
    def open_full(self):
        pass  # pragma: no cover

    @m.state()
    def closing():
        pass  # pragma: no cover

    @m.state()
    def closed():
        pass  # pragma: no cover

    @m.input()
    def connect_protocol_half(self):
        pass

    @m.input()
    def connect_protocol_full(self):
        pass

    @m.input()
    def remote_data(self, data):
        pass

    @m.input()
    def remote_close(self):
        pass

    @m.input()
    def local_data(self, data):
        pass

    @m.input()
    def local_close(self):
        pass

    @m.output()
    def queue_remote_data(self, data):
        self._pending_remote_data.append(data)

    @m.output()
    def queue_remote_close(self):
        self._pending_remote_close = True

    @m.output()
    def send_data(self, data):
        self._manager.send_data(self._scid, data)

    @m.output()
    def send_close(self):
        self._manager.send_close(self._scid)

    @m.output()
    def signal_dataReceived(self, data):
        assert self._protocol
        self._protocol.dataReceived(data)

    @m.output()
    def signal_readConnectionLost(self):
        IHalfCloseableProtocol(self._protocol).readConnectionLost()

    @m.output()
    def signal_writeConnectionLost(self):
        IHalfCloseableProtocol(self._protocol).writeConnectionLost()

    @m.output()
    def signal_connectionLost(self):
        assert self._protocol
        self._protocol.connectionLost(ConnectionDone())

    @m.output()
    def close_subchannel(self):
        self._manager.subchannel_closed(self._scid, self)
        # we're deleted momentarily

    @m.output()
    def error_closed_write(self, data):
        raise AlreadyClosedError("write not allowed on closed subchannel")

    @m.output()
    def error_closed_close(self):
        raise AlreadyClosedError(
            "loseConnection not allowed on closed subchannel")

    # stuff that arrives before we have a protocol connected
    unconnected.upon(remote_data, enter=unconnected, outputs=[queue_remote_data])
    unconnected.upon(remote_close, enter=unconnected, outputs=[queue_remote_close])

    # IHalfCloseableProtocol flow
    unconnected.upon(connect_protocol_half, enter=open_half, outputs=[])
    open_half.upon(remote_data, enter=open_half, outputs=[signal_dataReceived])
    open_half.upon(local_data, enter=open_half, outputs=[send_data])
    # remote closes first
    open_half.upon(remote_close, enter=read_closed, outputs=[signal_readConnectionLost])
    read_closed.upon(local_data, enter=read_closed, outputs=[send_data])
    read_closed.upon(local_close, enter=closed, outputs=[send_close,
                                                         close_subchannel,
                                                         # TODO: eventual-signal this?
                                                         signal_writeConnectionLost,
                                                         ])
    # local closes first
    open_half.upon(local_close, enter=write_closed, outputs=[signal_writeConnectionLost,
                                                             send_close])
    write_closed.upon(local_data, enter=write_closed, outputs=[error_closed_write])
    write_closed.upon(remote_data, enter=write_closed, outputs=[signal_dataReceived])
    write_closed.upon(remote_close, enter=closed, outputs=[close_subchannel,
                                                           signal_readConnectionLost,
                                                           ])
    # error cases
    write_closed.upon(local_close, enter=write_closed, outputs=[error_closed_close])

    # fully-closeable-only flow
    unconnected.upon(connect_protocol_full, enter=open_full, outputs=[])
    open_full.upon(remote_data, enter=open_full, outputs=[signal_dataReceived])
    open_full.upon(local_data, enter=open_full, outputs=[send_data])
    open_full.upon(remote_close, enter=closed, outputs=[send_close,
                                                        close_subchannel,
                                                        signal_connectionLost])
    open_full.upon(local_close, enter=closing, outputs=[send_close])
    closing.upon(remote_data, enter=closing, outputs=[signal_dataReceived])
    closing.upon(remote_close, enter=closed, outputs=[close_subchannel,
                                                      signal_connectionLost])
    # error cases
    # we won't ever see an OPEN, since L4 will log+ignore those for us
    closing.upon(local_data, enter=closing, outputs=[error_closed_write])
    closing.upon(local_close, enter=closing, outputs=[error_closed_close])
    # the CLOSED state shouldn't ever see messages, since we'll be deleted
    # (but a local user should be able to call "close" without having
    # to know what state we're in)
    closed.upon(local_close, enter=closed, outputs=[])

    # our endpoints use these

    def _set_protocol(self, protocol):
        assert not self._protocol
        self._protocol = protocol
        if IHalfCloseableProtocol.providedBy(protocol):
            self.connect_protocol_half()
        else:
            # move from UNCONNECTED to OPEN
            self.connect_protocol_full()

    def _deliver_queued_data(self):
        for data in self._pending_remote_data:
            self.remote_data(data)
        del self._pending_remote_data
        if self._pending_remote_close:
            self.remote_close()
            del self._pending_remote_close

    # ITransport
    def write(self, data):
        assert isinstance(data, bytes)
        assert len(data) <= MAX_FRAME_LENGTH
        self.local_data(data)

    def writeSequence(self, iovec):
        self.write(b"".join(iovec))

    def loseWriteConnection(self):
        if not IHalfCloseableProtocol.providedBy(self._protocol):
            # this is a clear error
            raise HalfCloseUsedOnNonHalfCloseable()
        self.local_close()

    def loseConnection(self):
        # TODO: what happens if an IHalfCloseableProtocol calls normal
        # loseConnection()? I think we need to close the read side too.
        if IHalfCloseableProtocol.providedBy(self._protocol):
            # I don't know is correct, so avoid this for now
            raise NormalCloseUsedOnHalfCloseable()
        self.local_close()

    def getHost(self):
        # we define "host addr" as the overall wormhole
        return self._host_addr

    def getPeer(self):
        # and "peer addr" as the subchannel within that wormhole
        return self._peer_addr

    # IProducer: throttle inbound data (wormhole "up" to local app's Protocol)
    def stopProducing(self):
        self._manager.subchannel_stopProducing(self)

    def pauseProducing(self):
        self._manager.subchannel_pauseProducing(self)

    def resumeProducing(self):
        self._manager.subchannel_resumeProducing(self)

    # IConsumer: allow the wormhole to throttle outbound data (app->wormhole)
    def registerProducer(self, producer, streaming):
        self._manager.subchannel_registerProducer(self, producer, streaming)

    def unregisterProducer(self):
        self._manager.subchannel_unregisterProducer(self)


@implementer(IStreamClientEndpoint)
@attrs
class SubchannelConnectorEndpoint:
    _subprotocol = attrib(validator=instance_of(str))
    _manager = attrib(validator=provides(IDilationManager))
    _host_addr = attrib(validator=instance_of(_WormholeAddress))
    _eventual_queue = attrib(repr=False)

    def __attrs_post_init__(self):
        self._connection_deferreds = deque()
        if not self._subprotocol:
            raise ValueError(
                "subprotocol must be a non-empty str"
            )

    @inlineCallbacks
    def connect(self, protocolFactory):
        # return Deferred that fires with IProtocol or Failure(ConnectError)
        yield self._manager._main_channel.when_fired()
        scid = self._manager.allocate_subchannel_id()
        self._manager.send_open(scid, self._subprotocol)
        peer_addr = SubchannelAddress(self._subprotocol)
        # ? f.doStart()
        # ? f.startedConnecting(CONNECTOR) # ??
        sc = SubChannel(scid, self._manager, self._host_addr, peer_addr)
        self._manager.subchannel_local_open(scid, sc)
        p = protocolFactory.buildProtocol(peer_addr)
        sc._set_protocol(p)
        p.makeConnection(sc)  # set p.transport = sc and call connectionMade()
        return p


class IllegalSubprotocolError(Exception):
    """
    A peer tried to open a subprotocol that has no listener.
    """


@implementer(IStreamServerEndpoint)
@attrs
class SubchannelListenerEndpoint:
    """
    This endpoint is used by application code to attach a factory to a
    given subprotocol. Instances are gotten via DilatedWormhole.listener_for()

    On listen(), we register for incoming OPENs on for this
    subprotocol name.
    """

    subprotocol_name = attrib()
    _manager = attrib()

    # this can, in fact, be async
    @inlineCallbacks
    def listen(self, factory):
        yield self._manager._main_channel.when_fired()
        self._manager._register_subprotocol_factory(self.subprotocol_name, factory)
        return SubchannelListeningPort(self._manager._host_addr)


class SubchannelDemultiplex:
    """
    Helper for Inbound to await factories for particular subprotocols,
    and deliver pending and future OPEN messages to them.

    If expected_subprotocols is `None` (the default) then all incoming
    OPENs are held pending a future listen. Otherwise,
    `expected_subprotocols` is a collection of str, where any OPEN not
    in the list produces an error.
    """
    def __init__(self, expected_subprotocols=None):
        self._factories = dict()  # name -> IProtocolFactory
        self._pending_opens = defaultdict(deque)  # name -> deque[tuple[transport, address]]
        self._expected = expected_subprotocols

    # from manager (actually Inbound)
    # t is Subchannel (transport) instance
    # peer_addr is a SubchannelAddress
    def _got_open(self, t, peer_addr):
        # t is "ITransport"
        name = peer_addr.subprotocol
        if name in self._factories:
            self._connect(self._factories[name], t, peer_addr)
        else:
            if self._expected is not None and name not in self._expected:
                raise UnexpectedSubprotocol()
            self._pending_opens[name].append((t, peer_addr))

    def _connect(self, factory, t, peer_addr):
        p = factory.buildProtocol(peer_addr)
        t._set_protocol(p)
        p.makeConnection(t)
        t._deliver_queued_data()

    def register(self, subprotocol_name, factory):
        if subprotocol_name in self._factories:
            raise ValueError(
                f'Already listening for subprotocol "{subprotocol_name}"'
            )
        self._factories[subprotocol_name] = factory

        # deliver any pending OPENs that have accumulated for this
        # subprotocol
        try:
            pending = self._pending_opens.pop(subprotocol_name)
        except KeyError:
            pending = deque()

        while pending:
            (t, peer_addr) = pending.popleft()
            self._connect(factory, t, peer_addr)


@implementer(IListeningPort)
@attrs
class SubchannelListeningPort:
    _host_addr = attrib(validator=provides(IAddress))

    def startListening(self):
        pass

    def stopListening(self):
        # TODO
        pass

    def getHost(self):
        return self._host_addr



from twisted.internet import defer

from pytest_twisted import ensureDeferred

from .. import xfer_util


APPID = "appid"


@ensureDeferred
async def test_xfer(reactor, mailbox):
    code = "1-code"
    data = "data"
    d1 = xfer_util.send(reactor, APPID, mailbox.url, data, code)
    d2 = xfer_util.receive(reactor, APPID, mailbox.url, code)
    send_result = await d1
    receive_result = await d2
    assert send_result is None
    assert receive_result == data


@ensureDeferred
async def test_on_code(reactor, mailbox):
    code = "1-code"
    data = "data"
    send_code = []
    receive_code = []
    d1 = xfer_util.send(
        reactor,
        APPID,
        mailbox.url,
        data,
        code,
        on_code=send_code.append)
    d2 = xfer_util.receive(
        reactor, APPID, mailbox.url, code, on_code=receive_code.append)
    send_result = await d1
    receive_result = await d2
    assert send_code == [code]
    assert receive_code == [code]
    assert send_result is None
    assert receive_result == data


@ensureDeferred
async def test_make_code(reactor, mailbox):
    data = "data"
    got_code = defer.Deferred()
    d1 = xfer_util.send(
        reactor,
        APPID,
        mailbox.url,
        data,
        code=None,
        on_code=got_code.callback)
    code = await got_code
    d2 = xfer_util.receive(reactor, APPID, mailbox.url, code)
    send_result = await d1
    receive_result = await d2
    assert send_result is None
    assert receive_result == data



from collections import defaultdict
from binascii import hexlify
from attr import attrs, attrib
from attr.validators import instance_of, optional
from automat import MethodicalMachine
from zope.interface import implementer
from twisted.internet.task import deferLater
from twisted.internet.defer import DeferredList, CancelledError
from twisted.internet.endpoints import serverFromString
from twisted.internet.protocol import ClientFactory, ServerFactory
from twisted.internet.address import HostnameAddress, IPv4Address, IPv6Address
from twisted.internet.error import ConnectingCancelledError, ConnectionRefusedError, DNSLookupError, ConnectError
from twisted.python import log
from .. import ipaddrs  # TODO: move into _dilation/
from .._interfaces import IDilationConnector, IDilationManager
from ..timing import DebugTiming
from ..observer import EmptyableSet
from ..util import HKDF, to_unicode, provides
from .connection import DilatedConnectionProtocol, KCM
from .roles import LEADER

from .._hints import (DirectTCPV1Hint, TorTCPV1Hint, RelayV1Hint,
                      parse_hint_argv, describe_hint_obj, endpoint_from_hint_obj,
                      encode_hint)
from .._status import DilationHint
from ._noise import NoiseConnection


def build_sided_relay_handshake(key, side):
    assert isinstance(side, str)
    # magic-wormhole-transit-relay expects a specific layout for the
    # handshake message: "please relay {64} for side {16}\n"
    assert len(side) == 8 * 2, side
    token = HKDF(key, 32, CTXinfo=b"transit_relay_token")
    return (b"please relay " + hexlify(token) +
            b" for side " + side.encode("ascii") + b"\n")


PROLOGUE_LEADER = b"Magic-Wormhole Dilation Handshake v1 Leader\n\n"
PROLOGUE_FOLLOWER = b"Magic-Wormhole Dilation Handshake v1 Follower\n\n"
NOISEPROTO = b"Noise_NNpsk0_25519_ChaChaPoly_BLAKE2s"


def build_noise():
    return NoiseConnection.from_name(NOISEPROTO)


@attrs(eq=False)
@implementer(IDilationConnector)
class Connector:
    """I manage a single generation of connection.

    The Manager creates one of me at a time, whenever it wants a connection
    (which is always, once w.dilate() has been called and we know the remote
    end can dilate, and is expressed by the Manager calling my .start()
    method). I am discarded when my established connection is lost (and if we
    still want to be connected, a new generation is started and a new
    Connector is created). I am also discarded if we stop wanting to be
    connected (which the Manager expresses by calling my .stop() method).

    I manage the race between multiple connections for a specific generation
    of the dilated connection.

    I send connection hints when my InboundConnectionFactory yields addresses
    (self.listener_ready), and I initiate outbond connections (with
    OutboundConnectionFactory) as I receive connection hints from my peer
    (self.got_hints). Both factories use my build_protocol() method to create
    connection.DilatedConnectionProtocol instances. I track these protocol
    instances until one finishes negotiation and wins the race. I then shut
    down the others, remember the winner as self._winning_connection, and
    deliver the winner to manager.connector_connection_made(c).

    When an active connection is lost, we call manager.connector_connection_lost,
    allowing the manager to decide whether it wants to start a new generation
    or not.
    """

    _dilation_key = attrib(validator=instance_of(bytes))
    _transit_relay_location = attrib(validator=optional(instance_of(str)))
    _manager = attrib(validator=provides(IDilationManager))
    _reactor = attrib()
    _eventual_queue = attrib()
    _no_listen = attrib(validator=instance_of(bool))
    _tor = attrib()
    _timing = attrib()
    _side = attrib(validator=instance_of(str))
    # was self._side = bytes_to_hexstr(os.urandom(8)) # unicode
    _role = attrib()

    m = MethodicalMachine()
    set_trace = getattr(m, "_setTrace", lambda self, f: None)  # pragma: no cover

    RELAY_DELAY = 2.0

    def __attrs_post_init__(self):
        if self._transit_relay_location:
            # TODO: allow multiple hints for a single relay
            relay_hint = parse_hint_argv(self._transit_relay_location)
            relay = RelayV1Hint(hints=(relay_hint,))
            self._transit_relays = [relay]
        else:
            self._transit_relays = []
        self._listeners = set()  # IListeningPorts that can be stopped
        self._pending_connectors = set()  # Deferreds that can be cancelled
        self._pending_connections = EmptyableSet(
            _eventual_queue=self._eventual_queue)  # Protocols to be stopped
        self._contenders = set()  # viable connections
        self._winning_connection = None
        self._timing = self._timing or DebugTiming()
        self._timing.add("transit")

    # this describes what our Connector can do, for the initial advertisement
    @classmethod
    def get_connection_abilities(klass):
        return [{"type": "direct-tcp-v1"},
                {"type": "relay-v1"},
                ]

    def build_protocol(self, addr, description):
        # encryption: let's use Noise NNpsk0 (or maybe NNpsk2). That uses
        # ephemeral keys plus a pre-shared symmetric key (the Transit key), a
        # different one for each potential connection.
        noise = build_noise()
        noise.set_psks(self._dilation_key)
        if self._role is LEADER:
            noise.set_as_initiator()
            outbound_prologue = PROLOGUE_LEADER
            inbound_prologue = PROLOGUE_FOLLOWER
        else:
            noise.set_as_responder()
            outbound_prologue = PROLOGUE_FOLLOWER
            inbound_prologue = PROLOGUE_LEADER
        p = DilatedConnectionProtocol(self._eventual_queue, self._role,
                                      description,
                                      self, noise,
                                      outbound_prologue, inbound_prologue)
        return p

    @m.state(initial=True)
    def connecting(self):
        pass  # pragma: no cover

    @m.state()
    def connected(self):
        pass  # pragma: no cover

    @m.state(terminal=True)
    def stopped(self):
        pass  # pragma: no cover

    # TODO: unify the tense of these method-name verbs

    @m.input()
    def got_hints(self, hint_objs):
        pass

    @m.input()
    def stop(self):
        pass

    # called by ourselves, when _start_listener() is ready
    @m.input()
    def listener_ready(self, hint_objs):
        pass

    # called when DilatedConnectionProtocol submits itself, after KCM
    # received
    @m.input()
    def add_candidate(self, c):
        pass

    # called by ourselves, via consider()
    @m.input()
    def accept(self, c):
        pass

    @m.output()
    def use_hints(self, hint_objs):
        self._use_hints(hint_objs)

    @m.output()
    def publish_hints(self, hint_objs):
        self._publish_hints(hint_objs)

    def _publish_hints(self, hint_objs):
        self._manager.send_hints([encode_hint(h) for h in hint_objs])

    @m.output()
    def consider(self, c):
        self._contenders.add(c)
        if self._role is LEADER:
            # for now, just accept the first one. TODO: be clever.
            self._eventual_queue.eventually(self.accept, c)
        else:
            # the follower always uses the first contender, since that's the
            # only one the leader picked
            self._eventual_queue.eventually(self.accept, c)

    @m.output()
    def select_and_stop_remaining(self, c):
        self._winning_connection = c
        self._contenders.clear()  # we no longer care who else came close
        # remove this winner from the losers, so we don't shut it down
        self._pending_connections.discard(c)
        # shut down losing connections
        self.stop_listeners()  # TODO: maybe keep it open? NAT/p2p assist
        self.stop_pending_connectors()
        self.stop_pending_connections()

        c.select(self._manager)  # subsequent frames go directly to the manager
        # c.select also wires up when_disconnected() to fire
        # manager.connector_connection_lost(). TODO: rename this, since the
        # Connector is no longer the one calling it
        if self._role is LEADER:
            # TODO: this should live in Connection
            c.send_record(KCM())  # leader sends KCM now
        self._manager.connector_connection_made(c)  # manager sends frames to Connection

    @m.output()
    def stop_everything(self):
        self.stop_listeners()
        self.stop_pending_connectors()
        self.stop_pending_connections()
        self.break_cycles()

    def stop_listeners(self):
        d = DeferredList([sub.stopListening() for sub in self._listeners])
        self._listeners.clear()
        return d  # synchronization for tests

    def stop_pending_connectors(self):
        for d in self._pending_connectors:
            d.cancel()

    def stop_pending_connections(self):
        d = self._pending_connections.when_next_empty()
        [c.disconnect() for c in self._pending_connections]
        return d

    def break_cycles(self):
        # help GC by forgetting references to things that reference us
        self._listeners.clear()
        self._pending_connectors.clear()
        self._pending_connections.clear()
        self._winning_connection = None

    connecting.upon(listener_ready, enter=connecting, outputs=[publish_hints])
    connecting.upon(got_hints, enter=connecting, outputs=[use_hints])
    connecting.upon(add_candidate, enter=connecting, outputs=[consider])
    connecting.upon(accept, enter=connected, outputs=[
                    select_and_stop_remaining])
    connecting.upon(stop, enter=stopped, outputs=[stop_everything])

    # once connected, we ignore everything except stop
    connected.upon(listener_ready, enter=connected, outputs=[])
    connected.upon(got_hints, enter=connected, outputs=[])
    # TODO: tell them to disconnect? will they hang out forever? I *think*
    # they'll drop this once they get a KCM on the winning connection.
    connected.upon(add_candidate, enter=connected, outputs=[])
    connected.upon(accept, enter=connected, outputs=[])
    connected.upon(stop, enter=stopped, outputs=[stop_everything])

    # from Manager: start, got_hints, stop
    # maybe add_candidate, accept

    def start(self):
        if not self._no_listen and not self._tor:
            addresses = self._get_listener_addresses()
            self._start_listener(addresses)
        if self._transit_relays:
            self._publish_hints(self._transit_relays)
            self._use_hints(self._transit_relays)

    def _get_listener_addresses(self):
        addresses = ipaddrs.find_addresses()
        non_loopback_addresses = [a for a in addresses if a != "127.0.0.1"]
        if non_loopback_addresses:
            # some test hosts, including the appveyor VMs, *only* have
            # 127.0.0.1, and the tests will hang badly if we remove it.
            addresses = non_loopback_addresses
        return addresses

    def _start_listener(self, addresses):
        # TODO: listen on a fixed port, if possible, for NAT/p2p benefits, also
        # to make firewall configs easier
        # TODO: retain listening port between connection generations?
        ep = serverFromString(self._reactor, "tcp:0")
        f = InboundConnectionFactory(self)
        d = ep.listen(f)

        def _listening(lp):
            # lp is an IListeningPort
            self._listeners.add(lp)  # for shutdown and tests
            portnum = lp.getHost().port
            direct_hints = [DirectTCPV1Hint(to_unicode(addr), portnum, 0.0)
                            for addr in addresses]
            self.listener_ready(direct_hints)
        d.addCallback(_listening)
        d.addErrback(log.err)

    def _schedule_connection(self, delay, h, is_relay):
        ep = endpoint_from_hint_obj(h, self._tor, self._reactor)
        desc = describe_hint_obj(h, is_relay, self._tor)
        d = deferLater(self._reactor, delay,
                       self._connect, ep, desc, is_relay)

        # "ConnectError" is a base-class in Twisted, but can be raised
        # directly when the "errno to class" mapping doesn't have an
        # error-number mapped
        d.addErrback(lambda f: f.trap(ConnectingCancelledError,
                                      ConnectionRefusedError,
                                      CancelledError,
                                      ConnectError,
                                      ))
        # TODO: HostnameEndpoint.connect catches CancelledError and replaces
        # it with DNSLookupError. Remove this workaround when
        # https://twistedmatrix.com/trac/ticket/9696 is fixed.
        d.addErrback(lambda f: f.trap(DNSLookupError))
        d.addErrback(log.err)
        self._pending_connectors.add(d)

    def _use_hints(self, hints):
        # first, pull out all the relays, we'll connect to them later
        relays = []
        direct = defaultdict(list)
        hint_status = []
        for h in hints:
            if isinstance(h, RelayV1Hint):
                relays.append(h)
            else:
                direct[h.priority].append(h)
        delay = 0.0
        made_direct = False
        priorities = sorted(set(direct.keys()), reverse=True)
        for p in priorities:
            for h in direct[p]:
                if isinstance(h, TorTCPV1Hint) and not self._tor:
                    continue
                hint_status.append(DilationHint(f"{h.hostname}:{h.port}", True))
                self._schedule_connection(delay, h, is_relay=False)
                made_direct = True
                # Make all direct connections immediately. Later, we'll change
                # the add_candidate() function to look at the priority when
                # deciding whether to accept a successful connection or not,
                # and it can wait for more options if it sees a higher-priority
                # one still running. But if we bail on that, we might consider
                # putting an inter-direct-hint delay here to influence the
                # process.
                # delay += 1.0

        if made_direct and not self._no_listen:
            # Prefer direct connections by stalling relay connections by a
            # few seconds. We don't wait until direct connections have
            # failed, because many direct hints will be to unused
            # local-network IP address, which won't answer, and can take the
            # full 30s TCP timeout to fail.
            #
            # If we didn't make any direct connections, or we're using
            # --no-listen, then we're probably going to have to use the
            # relay, so don't delay it at all.
            delay += self.RELAY_DELAY

        # It might be nice to wire this so that a failure in the direct hints
        # causes the relay hints to be used right away (fast failover). But
        # none of our current use cases would take advantage of that: if we
        # have any viable direct hints, then they're either going to succeed
        # quickly or hang for a long time.
        for r in relays:
            for h in r.hints:
                self._schedule_connection(delay, h, is_relay=True)
                hint_status.append(DilationHint(f"{h.hostname}:{h.port}", False))

        self._manager._hint_status(hint_status)
        # TODO:
        # if not contenders:
        #    raise TransitError("No contenders for connection")

    # TODO: add 2*TIMEOUT deadline for first generation, don't wait forever for
    # the initial connection

    def _connect(self, ep, description, is_relay=False):
        relay_handshake = None
        if is_relay:
            relay_handshake = build_sided_relay_handshake(self._dilation_key,
                                                          self._side)
        f = OutboundConnectionFactory(self, relay_handshake, description)
        d = ep.connect(f)
        # fires with protocol, or ConnectError

        def _connected(p):
            self._pending_connections.add(p)
            # c might not be in _pending_connections, if it turned out to be a
            # winner, which is why we use discard() and not remove()
            p.when_disconnected().addCallback(self._pending_connections.discard)
        d.addCallback(_connected)
        return d

    # Connection selection. All instances of DilatedConnectionProtocol which
    # look viable get passed into our add_contender() method.

    # On the Leader side, "viable" means we've seen their KCM frame, which is
    # the first Noise-encrypted packet on any given connection, and it has an
    # empty body. We gather viable connections until we see one that we like,
    # or a timer expires. Then we "select" it, close the others, and tell our
    # Manager to use it.

    # On the Follower side, we'll only see a KCM on the one connection selected
    # by the Leader, so the first viable connection wins.

    # our Connection protocols call: add_candidate


@attrs(repr=False)
class OutboundConnectionFactory(ClientFactory):
    _connector = attrib(validator=provides(IDilationConnector))
    _relay_handshake = attrib(validator=optional(instance_of(bytes)))
    _description = attrib()

    def __repr__(self):
        return f"OutboundConnectionFactory({self._connector._role} {self._description})"

    def buildProtocol(self, addr):
        p = self._connector.build_protocol(addr, self._description)
        p.factory = self
        if self._relay_handshake is not None:
            p.use_relay(self._relay_handshake)
        return p


def describe_inbound(addr):
    if isinstance(addr, HostnameAddress):
        return "<-tcp:%s:%d" % (addr.hostname, addr.port)
    elif isinstance(addr, IPv4Address):
        return "<-tcp:%s:%d" % (addr.host, addr.port)
    elif isinstance(addr, IPv6Address):
        return "<-tcp:[%s]:%d" % (addr.host, addr.port)
    return f"<-{addr!r}"


@attrs(repr=False)
class InboundConnectionFactory(ServerFactory):
    _connector = attrib(validator=provides(IDilationConnector))

    def __repr__(self):
        return f"InboundConnectionFactory({self._connector._role})"

    def buildProtocol(self, addr):
        description = describe_inbound(addr)
        p = self._connector.build_protocol(addr, description)
        p.factory = self
        return p



from zope.interface import alsoProvides
from twisted.internet.task import Clock, Cooperator
from twisted.internet.interfaces import IStreamServerEndpoint
from unittest import mock
import pytest
import pytest_twisted

from ...eventual import EventualQueue
from ..._interfaces import ISend, ITerminator, ISubChannel
from ...util import dict_to_bytes
from ..._dilation import roles
from ..._dilation.manager import (Dilator, Manager, make_side,
                                  OldPeerCannotDilateError,
                                  CanOnlyDilateOnceError,
                                  UnknownDilationMessageType,
                                  UnexpectedKCM,
                                  UnknownMessageType, DILATION_VERSIONS)
from ..._dilation.connection import Open, Data, Close, Ack, KCM, Ping, Pong
from .common import clear_mock_calls


class Holder():
    pass


def make_dilator():
    h = Holder()
    h.reactor = object()
    h.clock = Clock()
    h.eq = EventualQueue(h.clock)
    term = mock.Mock(side_effect=lambda: True)  # one write per Eventual tick

    def term_factory():
        return term
    h.coop = Cooperator(terminationPredicateFactory=term_factory,
                        scheduler=h.eq.eventually)
    h.send = mock.Mock()
    alsoProvides(h.send, ISend)
    dil = Dilator(h.reactor, h.eq, h.coop, DILATION_VERSIONS)
    h.terminator = mock.Mock()
    alsoProvides(h.terminator, ITerminator)
    dil.wire(h.send, h.terminator)
    return dil, h


# we should test the interleavings between:
# * application calls w.dilate() and gets back endpoints
# * wormhole gets: dilation key, VERSION, 0-n dilation messages


def test_dilate_first():
    (dil, h) = make_dilator()
    side = object()
    m = mock.Mock()
    m._api = eps = object()
    mm = mock.Mock(side_effect=[m])
    with mock.patch("wormhole._dilation.manager.Manager", mm), \
         mock.patch("wormhole._dilation.manager.make_side",
                    return_value=side):
        eps1 = dil.dilate()
        with pytest.raises(CanOnlyDilateOnceError):
            dil.dilate()
    assert eps1 is eps
    assert mm.mock_calls == [mock.call(h.send, side, None,
                                       h.reactor, h.eq, h.coop, DILATION_VERSIONS, 30.0, None,
                                       False, None, initial_mailbox_status=None)]

    assert m.mock_calls == []

    key = b"key"
    transit_key = object()
    with mock.patch("wormhole._dilation.manager.derive_key",
                    return_value=transit_key) as dk:
        dil.got_key(key)
    assert dk.mock_calls == [mock.call(key, b"dilation-v1", 32)]
    assert m.mock_calls == [mock.call.got_dilation_key(transit_key)]
    clear_mock_calls(m)

    wv = object()
    dil.got_wormhole_versions(wv)
    assert m.mock_calls == [mock.call.got_wormhole_versions(wv)]
    clear_mock_calls(m)

    dm1 = object()
    dm2 = object()
    dil.received_dilate(dm1)
    dil.received_dilate(dm2)
    assert m.mock_calls == [mock.call.received_dilation_message(dm1),
                                    mock.call.received_dilation_message(dm2),
                                    ]
    clear_mock_calls(m)

    stopped_d = mock.Mock()
    m.when_stopped = mock.Mock(return_value=stopped_d)
    dil.stop()
    assert m.mock_calls == [mock.call.stop(),
                                    mock.call.when_stopped(),
                                    ]


def test_dilate_later():
    (dil, h) = make_dilator()
    m = mock.Mock()
    mm = mock.Mock(side_effect=[m])

    key = b"key"
    transit_key = object()
    with mock.patch("wormhole._dilation.manager.derive_key",
                    return_value=transit_key) as dk:
        dil.got_key(key)
    assert dk.mock_calls == [mock.call(key, b"dilation-v1", 32)]

    wv = object()
    dil.got_wormhole_versions(wv)

    dm1 = object()
    dil.received_dilate(dm1)

    assert mm.mock_calls == []

    with mock.patch("wormhole._dilation.manager.Manager", mm):
        dil.dilate()
    assert m.mock_calls == [mock.call.got_dilation_key(transit_key),
                                    mock.call.got_wormhole_versions(wv),
                                    mock.call.received_dilation_message(dm1),
                                    ]
    clear_mock_calls(m)

    dm2 = object()
    dil.received_dilate(dm2)
    assert m.mock_calls == [mock.call.received_dilation_message(dm2),
                                    ]

def test_stop_early():
    (dil, h) = make_dilator()
    # we stop before w.dilate(), so there is no Manager to stop
    dil.stop()
    assert h.terminator.mock_calls == [mock.call.stoppedD()]


@pytest_twisted.ensureDeferred
async def test_peer_cannot_dilate():
    (dil, h) = make_dilator()
    eps = dil.dilate()

    dil.got_key(b"\x01" * 32)
    dil.got_wormhole_versions({})  # missing "can-dilate"
    f = mock.Mock()
    d = eps.connector_for("proto").connect(f)
    h.eq.flush_sync()
    with pytest.raises(OldPeerCannotDilateError):
        await d


@pytest_twisted.ensureDeferred
async def test_disjoint_versions():
    (dil, h) = make_dilator()
    eps = dil.dilate()

    dil.got_key(b"\x01" * 32)
    dil.got_wormhole_versions({"can-dilate": ["-1"]})
    f = mock.Mock()
    d = eps.connector_for("proto").connect(f)
    h.eq.flush_sync()
    with pytest.raises(OldPeerCannotDilateError):
        await d


def test_transit_relay():
    (dil, h) = make_dilator()
    transit_relay_location = object()
    side = object()
    m = mock.Mock()
    mm = mock.Mock(side_effect=[m])
    with mock.patch("wormhole._dilation.manager.Manager", mm), \
         mock.patch("wormhole._dilation.manager.make_side",
                    return_value=side):
        dil.dilate(transit_relay_location)
    assert mm.mock_calls == [mock.call(h.send, side, transit_relay_location,
                                       h.reactor, h.eq, h.coop, DILATION_VERSIONS, 30.0, None,
                                       False, None, initial_mailbox_status=None)]


LEADER = "ff3456abcdef"
FOLLOWER = "123456abcdef"


class FakeHost:
    def __init__(self, port):
        self.port = port


class FakePort:
    def __init__(self, port):
        self.port = port

    def getHost(self):
        return FakeHost(self.port)


class ReactorFake:
    """
    Provide a reactor-like mock
    """
    # not ideal, but prior to this the "reactor" in these tests was
    # literally just "object()"

    def __init__(self):
        self.listens = []

    def seconds(self):
        return 42

    def callLater(self, interval, callable, *args):
        # should return a DelayedCall
        return mock.Mock()

    def listenTCP(self, *args, **kw):
        # returns an IListenPort
        self.listens.append((args, kw))
        return FakePort(1234)


def make_manager(leader=True):
    h = Holder()
    h.send = mock.Mock()
    alsoProvides(h.send, ISend)
    if leader:
        side = LEADER
    else:
        side = FOLLOWER
    h.key = b"\x00" * 32
    h.relay = None
    h.reactor = ReactorFake()
    h.clock = Clock()
    h.eq = EventualQueue(h.clock)
    term = mock.Mock(side_effect=lambda: True)  # one write per Eventual tick

    def term_factory():
        return term
    h.coop = Cooperator(terminationPredicateFactory=term_factory,
                        scheduler=h.eq.eventually)
    h.inbound = mock.Mock()
    h.Inbound = mock.Mock(return_value=h.inbound)
    h.outbound = mock.Mock()
    h.Outbound = mock.Mock(return_value=h.outbound)
    h.sc0 = mock.Mock()
    alsoProvides(h.sc0, ISubChannel)
    h.SubChannel = mock.Mock(return_value=h.sc0)
    h.listen_ep = mock.Mock()
    alsoProvides(h.listen_ep, IStreamServerEndpoint)
    with mock.patch("wormhole._dilation.manager.Inbound", h.Inbound), \
         mock.patch("wormhole._dilation.manager.Outbound", h.Outbound), \
         mock.patch("wormhole._dilation.subchannel.SubChannel", h.SubChannel), \
         mock.patch("wormhole._dilation.manager.SubchannelListenerEndpoint",
                    return_value=h.listen_ep):
        m = Manager(h.send, side, h.relay, h.reactor, h.eq, h.coop, DILATION_VERSIONS, 30.0, {})
    h.hostaddr = m._host_addr
    m.got_dilation_key(h.key)
    return m, h


def test_make_side():
    side = make_side()
    assert type(side) is str
    assert len(side) == 2 * 8


def test_create():
    m, h = make_manager()


def test_leader():
    m, h = make_manager(leader=True)
    assert h.send.mock_calls == []
    assert h.Inbound.mock_calls == [mock.call(m, h.hostaddr)]
    assert h.Outbound.mock_calls == [mock.call(m, h.coop)]
    assert h.SubChannel.mock_calls == []
    assert h.inbound.mock_calls == []
    clear_mock_calls(h.inbound)

    m.got_wormhole_versions({"can-dilate": ["ged"]})
    assert h.send.mock_calls == [
        mock.call.send("dilate-0",
                       dict_to_bytes({"type": "please", "side": LEADER, "use-version": "ged"}))
        ]
    clear_mock_calls(h.send)

    # ignore early hints
    m.rx_HINTS({})
    assert h.send.mock_calls == []

    c = mock.Mock()
    connector = mock.Mock(return_value=c)
    with mock.patch("wormhole._dilation.manager.Connector", connector):
        # receiving this PLEASE triggers creation of the Connector
        m.rx_PLEASE({"side": FOLLOWER})
    assert h.send.mock_calls == []
    assert connector.mock_calls == [
        mock.call(b"\x00" * 32, None, m, h.reactor, h.eq,
                  False,  # no_listen
                  None,  # tor
                  None,  # timing
                  LEADER, roles.LEADER),
        ]
    assert c.mock_calls == [mock.call.start()]
    clear_mock_calls(connector, c)

    # now any inbound hints should get passed to our Connector
    with mock.patch("wormhole._dilation.manager.parse_hint",
                    side_effect=["p1", None, "p3"]) as ph:
        m.rx_HINTS({"hints": [1, 2, 3]})
    assert ph.mock_calls == [mock.call(1), mock.call(2), mock.call(3)]
    assert c.mock_calls == [mock.call.got_hints(["p1", "p3"])]
    clear_mock_calls(ph, c)

    # and we send out any (listening) hints from our Connector
    m.send_hints([1, 2])
    assert h.send.mock_calls == [
        mock.call.send("dilate-1",
                       dict_to_bytes({"type": "connection-hints",
                                      "hints": [1, 2]}))
        ]
    clear_mock_calls(h.send)

    # the first successful connection fires when_first_connected(), so
    # the endpoints can activate
    c1 = mock.Mock()
    m.connector_connection_made(c1)

    assert h.inbound.mock_calls == [mock.call.use_connection(c1)]
    assert h.outbound.mock_calls[1:] == [mock.call.use_connection(c1)]
    clear_mock_calls(h.inbound, h.outbound)

    # the Leader making a new outbound channel should get scid=1
    scid1 = 1
    assert m.allocate_subchannel_id() == scid1
    r1 = Open(10, scid1, "proto")  # seqnum=10
    h.outbound.build_record = mock.Mock(return_value=r1)
    m.send_open(scid1, "proto")
    assert h.outbound.mock_calls == [
        mock.call.build_record(Open, scid1, "proto"),
        mock.call.queue_and_send_record(r1),
        ]
    clear_mock_calls(h.outbound)

    r2 = Data(11, scid1, b"data")
    h.outbound.build_record = mock.Mock(return_value=r2)
    m.send_data(scid1, b"data")
    assert h.outbound.mock_calls == [
        mock.call.build_record(Data, scid1, b"data"),
        mock.call.queue_and_send_record(r2),
        ]
    clear_mock_calls(h.outbound)

    r3 = Close(12, scid1)
    h.outbound.build_record = mock.Mock(return_value=r3)
    m.send_close(scid1)
    assert h.outbound.mock_calls == [
        mock.call.build_record(Close, scid1),
        mock.call.queue_and_send_record(r3),
        ]
    clear_mock_calls(h.outbound)

    # ack the OPEN
    m.got_record(Ack(10))
    assert h.outbound.mock_calls == [
        mock.call.handle_ack(10)
        ]
    clear_mock_calls(h.outbound)

    # test that inbound records get acked and routed to Inbound
    h.inbound.is_record_old = mock.Mock(return_value=False)
    scid2 = 2
    o200 = Open(200, scid2, "proto")
    m.got_record(o200)
    assert h.outbound.mock_calls == [
        mock.call.send_if_connected(Ack(200))
        ]
    assert h.inbound.mock_calls == [
        mock.call.is_record_old(o200),
        mock.call.update_ack_watermark(200),
        mock.call.handle_open(scid2, "proto"),
        ]
    clear_mock_calls(h.outbound, h.inbound)

    # old (duplicate) records should provoke new Acks, but not get
    # forwarded
    h.inbound.is_record_old = mock.Mock(return_value=True)
    m.got_record(o200)
    assert h.outbound.mock_calls == [
        mock.call.send_if_connected(Ack(200))
        ]
    assert h.inbound.mock_calls == [
        mock.call.is_record_old(o200),
        ]
    clear_mock_calls(h.outbound, h.inbound)

    # check Data and Close too
    h.inbound.is_record_old = mock.Mock(return_value=False)
    d201 = Data(201, scid2, b"data")
    m.got_record(d201)
    assert h.outbound.mock_calls == [
        mock.call.send_if_connected(Ack(201))
        ]
    assert h.inbound.mock_calls == [
        mock.call.is_record_old(d201),
        mock.call.update_ack_watermark(201),
        mock.call.handle_data(scid2, b"data"),
        ]
    clear_mock_calls(h.outbound, h.inbound)

    c202 = Close(202, scid2)
    m.got_record(c202)
    assert h.outbound.mock_calls == [
        mock.call.send_if_connected(Ack(202))
        ]
    assert h.inbound.mock_calls == [
        mock.call.is_record_old(c202),
        mock.call.update_ack_watermark(202),
        mock.call.handle_close(scid2),
        ]
    clear_mock_calls(h.outbound, h.inbound)

    # Now we lose the connection. The Leader should tell the other side
    # that we're reconnecting.

    m.connector_connection_lost()
    assert h.send.mock_calls == [
        mock.call.send("dilate-2",
                       dict_to_bytes({"type": "reconnect"}))
        ]
    assert h.inbound.mock_calls == [
        mock.call.stop_using_connection()
        ]
    assert h.outbound.mock_calls == [
        mock.call.stop_using_connection()
        ]
    clear_mock_calls(h.send, h.inbound, h.outbound)

    # leader does nothing (stays in FLUSHING) until the follower acks by
    # sending RECONNECTING

    # inbound hints should be ignored during FLUSHING
    with mock.patch("wormhole._dilation.manager.parse_hint",
                    return_value=None) as ph:
        m.rx_HINTS({"hints": [1, 2, 3]})
    assert ph.mock_calls == []  # ignored

    c2 = mock.Mock()
    connector2 = mock.Mock(return_value=c2)
    with mock.patch("wormhole._dilation.manager.Connector", connector2):
        # this triggers creation of a new Connector
        m.rx_RECONNECTING()
    assert h.send.mock_calls == []
    assert connector2.mock_calls == [
        mock.call(b"\x00" * 32, None, m, h.reactor, h.eq,
                  False,  # no_listen
                  None,  # tor
                  None,  # timing
                  LEADER, roles.LEADER),
        ]
    assert c2.mock_calls == [mock.call.start()]
    clear_mock_calls(connector2, c2)

    assert h.inbound.mock_calls == []
    assert h.outbound.mock_calls == []

    # and a new connection should re-register with Inbound/Outbound,
    # which are responsible for re-sending unacked queued messages
    c3 = mock.Mock()
    m.connector_connection_made(c3)

    assert h.inbound.mock_calls == [mock.call.use_connection(c3)]
    assert h.outbound.mock_calls[1:] == [mock.call.use_connection(c3)]
    clear_mock_calls(h.inbound, h.outbound)


def test_follower():
    m, h = make_manager(leader=False)

    m.got_wormhole_versions({"can-dilate": ["ged"]})
    assert h.send.mock_calls == [
        mock.call.send("dilate-0",
                       dict_to_bytes({"type": "please", "side": FOLLOWER, "use-version": "ged"}))
        ]
    clear_mock_calls(h.send)
    clear_mock_calls(h.inbound)

    c = mock.Mock()
    connector = mock.Mock(return_value=c)
    with mock.patch("wormhole._dilation.manager.Connector", connector):
        # receiving this PLEASE triggers creation of the Connector
        m.rx_PLEASE({"side": LEADER})
    assert h.send.mock_calls == []
    assert connector.mock_calls == [
        mock.call(b"\x00" * 32, None, m, h.reactor, h.eq,
                  False,  # no_listen
                  None,  # tor
                  None,  # timing
                  FOLLOWER, roles.FOLLOWER),
        ]
    assert c.mock_calls == [mock.call.start()]
    clear_mock_calls(connector, c)

    # get connected, then lose the connection
    c1 = mock.Mock()
    m.connector_connection_made(c1)
    assert h.inbound.mock_calls == [mock.call.use_connection(c1)]
    assert h.outbound.mock_calls == [mock.call.use_connection(c1)]
    clear_mock_calls(h.inbound, h.outbound)

    # now lose the connection. As the follower, we don't notify the
    # leader, we just wait for them to notice
    m.connector_connection_lost()
    assert h.send.mock_calls == []
    assert h.inbound.mock_calls == [
        mock.call.stop_using_connection()
        ]
    assert h.outbound.mock_calls == [
        mock.call.stop_using_connection()
        ]
    clear_mock_calls(h.send, h.inbound, h.outbound)

    # now we get a RECONNECT: we should send RECONNECTING
    c2 = mock.Mock()
    connector2 = mock.Mock(return_value=c2)
    with mock.patch("wormhole._dilation.manager.Connector", connector2):
        m.rx_RECONNECT()
    assert h.send.mock_calls == [
        mock.call.send("dilate-1",
                       dict_to_bytes({"type": "reconnecting"}))
        ]
    assert connector2.mock_calls == [
        mock.call(b"\x00" * 32, None, m, h.reactor, h.eq,
                  False,  # no_listen
                  None,  # tor
                  None,  # timing
                  FOLLOWER, roles.FOLLOWER),
        ]
    assert c2.mock_calls == [mock.call.start()]
    clear_mock_calls(connector2, c2)

    # while we're trying to connect, we get told to stop again, so we
    # should abandon the connection attempt and start another
    c3 = mock.Mock()
    connector3 = mock.Mock(return_value=c3)
    with mock.patch("wormhole._dilation.manager.Connector", connector3):
        m.rx_RECONNECT()
    assert c2.mock_calls == [mock.call.stop()]
    assert connector3.mock_calls == [
        mock.call(b"\x00" * 32, None, m, h.reactor, h.eq,
                  False,  # no_listen
                  None,  # tor
                  None,  # timing
                  FOLLOWER, roles.FOLLOWER),
        ]
    assert c3.mock_calls == [mock.call.start()]
    clear_mock_calls(c2, connector3, c3)

    m.connector_connection_made(c3)
    # finally if we're already connected, rx_RECONNECT means we should
    # abandon this connection (even though it still looks ok to us), then
    # when the attempt is finished stopping, we should start another

    m.rx_RECONNECT()

    c4 = mock.Mock()
    connector4 = mock.Mock(return_value=c4)
    with mock.patch("wormhole._dilation.manager.Connector", connector4):
        m.connector_connection_lost()
    assert c3.mock_calls == [mock.call.disconnect()]
    assert connector4.mock_calls == [
        mock.call(b"\x00" * 32, None, m, h.reactor, h.eq,
                  False,  # no_listen
                  None,  # tor
                  None,  # timing
                  FOLLOWER, roles.FOLLOWER),
        ]
    assert c4.mock_calls == [mock.call.start()]
    clear_mock_calls(c3, connector4, c4)


def test_mirror():
    # receive a PLEASE with the same side as us: shouldn't happen
    m, h = make_manager(leader=True)

    m.start()
    clear_mock_calls(h.send)
    with pytest.raises(ValueError) as f:
        m.rx_PLEASE({"side": LEADER})
    assert str(f.value) == "their side shouldn't be equal: reflection?"


def test_ping_pong(observe_errors):
    m, h = make_manager(leader=False)

    m.got_record(KCM())
    observe_errors.flush(UnexpectedKCM)

    m.got_record(Ping(1))
    assert h.outbound.mock_calls == \
                     [mock.call.send_if_connected(Pong(1))]
    clear_mock_calls(h.outbound)

    m.got_record("not recognized")
    e = observe_errors.flush(UnknownMessageType)
    assert len(e) == 1
    assert str(e[0].value) == "not recognized"

    m.send_ping(2, lambda _: None)
    assert h.outbound.mock_calls == \
                     [mock.call.send_if_connected(Pong(2))]
    clear_mock_calls(h.outbound)

    # sort of low-level; what does this look like to API user?
    class FakeError(Exception):
        pass

    def cause_error(_):
        raise FakeError()
    m.send_ping(3, cause_error)
    assert h.outbound.mock_calls == \
                     [mock.call.send_if_connected(Pong(3))]
    clear_mock_calls(h.outbound)
    with pytest.raises(FakeError):
        m.got_record(Pong(3))


def test_subchannel():
    m, h = make_manager(leader=True)
    clear_mock_calls(h.inbound)
    sc = object()

    m.subchannel_pauseProducing(sc)
    assert h.inbound.mock_calls == [
        mock.call.subchannel_pauseProducing(sc)]
    clear_mock_calls(h.inbound)

    m.subchannel_resumeProducing(sc)
    assert h.inbound.mock_calls == [
        mock.call.subchannel_resumeProducing(sc)]
    clear_mock_calls(h.inbound)

    m.subchannel_stopProducing(sc)
    assert h.inbound.mock_calls == [
        mock.call.subchannel_stopProducing(sc)]
    clear_mock_calls(h.inbound)

    p = object()
    streaming = object()

    m.subchannel_registerProducer(sc, p, streaming)
    assert h.outbound.mock_calls == [
        mock.call.subchannel_registerProducer(sc, p, streaming)]
    clear_mock_calls(h.outbound)

    m.subchannel_unregisterProducer(sc)
    assert h.outbound.mock_calls == [
        mock.call.subchannel_unregisterProducer(sc)]
    clear_mock_calls(h.outbound)

    m.subchannel_closed(4, sc)
    assert h.inbound.mock_calls == [
        mock.call.subchannel_closed(4, sc)]
    assert h.outbound.mock_calls == [
        mock.call.subchannel_closed(4, sc)]
    clear_mock_calls(h.inbound, h.outbound)


def test_unknown_message(observe_errors):
    # receive a PLEASE with the same side as us: shouldn't happen
    m, h = make_manager(leader=True)
    m.start()

    m.received_dilation_message(dict_to_bytes(dict(type="unknown")))
    observe_errors.flush(UnknownDilationMessageType)


def test_status_hints():
    """
    Rudimentary test proving that sending two different hint updates
    results in a final update with both hints
    """
    m, h = make_manager()

    # better if we could use Dilator -> dilate() -> etc
    updates = []
    def status_update(st):
        updates.append(st)
    m._status = status_update

    m.start()
    m.rx_PLEASE({"side": "fake side"})
    m.rx_HINTS({"hints": [
        {
            "type": "direct-tcp-v1",
            "hostname": "fake.host",
            "port": 1,
        }
    ]})
    m.rx_HINTS({"hints": [
        {
            "type": "direct-tcp-v1",
            "hostname": "fake.host",
            "port": 2,
        }
    ]})

    # we've done two (separate!) hint updates, our last status update
    # should show both of them

    last_hints = [
        hint.url
        for hint in updates[-1].hints
    ]
    assert "fake.host:1" in last_hints
    assert "fake.host:2" in last_hints


# TODO: test transit relay is used
